def get_credential(
    credential_type,
    tenant_id,
    client_id,
    client_secret,
    certificate_path,
):
    credential_type = credential_type
    if credential_type == "default":
        from azure.identity import DefaultAzureCredential

        return DefaultAzureCredential()

    if credential_type == "client_secret":
        from azure.identity import ClientSecretCredential

        return ClientSecretCredential(
            tenant_id=tenant_id,
            client_id=client_id,
            client_secret=client_secret,
        )

    if credential_type == "certificate":
        from azure.identity import CertificateCredential

        return CertificateCredential(
            tenant_id=tenant_id,
            client_id=client_id,
            certificate_path=certificate_path,
        )

    if credential_type == "azure_cli":
        from azure.identity import AzureCliCredential

        return AzureCliCredential()

    if credential_type == "shared_token_cache":
        from azure.identity import SharedTokenCacheCredential

        return SharedTokenCacheCredential()

    if credential_type == "managed_identity":
        from azure.identity import ManagedIdentityCredential

        return ManagedIdentityCredential()
    return None


def aget_credential(
    credential_type,
    tenant_id,
    client_id,
    client_secret,
    certificate_path,
):
    credential_type = credential_type
    if credential_type == "default":
        from azure.identity.aio import DefaultAzureCredential

        return DefaultAzureCredential()

    if credential_type == "client_secret":
        from azure.identity.aio import ClientSecretCredential

        return ClientSecretCredential(
            tenant_id=tenant_id,
            client_id=client_id,
            client_secret=client_secret,
        )

    if credential_type == "certificate":
        from azure.identity.aio import CertificateCredential

        return CertificateCredential(
            tenant_id=tenant_id,
            client_id=client_id,
            certificate_path=certificate_path,
        )

    if credential_type == "azure_cli":
        from azure.identity.aio import AzureCliCredential

        return AzureCliCredential()

    if credential_type == "shared_token_cache":
        from azure.identity.aio import SharedTokenCacheCredential

        return SharedTokenCacheCredential()

    if credential_type == "managed_identity":
        from azure.identity.aio import ManagedIdentityCredential

        return ManagedIdentityCredential()
    return None



"""
Object Store on File System.
"""

from __future__ import annotations

__all__ = ["FileSystem"]

import os
from typing import Any

from verse.core import Context, NCall, Operation, Response
from verse.core.exceptions import BadRequestError
from verse.internal.storage_core import (
    ItemProcessor,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
)

from .._helper import get_collection_config
from .._models import ObjectCollectionConfig


class FileSystem(StoreProvider):
    store_path: str
    folder: str | None
    nparams: dict[str, Any]

    _init: bool
    _collection_cache: dict[str, FileSystemCollection]

    def __init__(
        self,
        store_path: str = ".",
        folder: str | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            store_path:
                Base store path. Defaults to ".".
            folder:
                Folder name mapped to object store collection name.
            nparams:
                Native parameters to file system operations.
        """
        self.store_path = store_path
        self.folder = folder
        self.nparams = nparams

        self._init = False
        self._collection_cache = dict()

    def init(self, context: Context | None = None) -> None:
        if not self._init:
            return

        self._init = True

    def _get_folder_name(self, op_parser: StoreOperationParser) -> str | None:
        collection_name = (
            op_parser.get_operation_parsers()[0].get_collection_name()
            if op_parser.op_equals(StoreOperation.BATCH)
            else op_parser.get_collection_name()
        )
        folder = (
            collection_name if collection_name is not None else self.folder
        )
        return folder

    def _get_collection(
        self, op_parser: StoreOperationParser
    ) -> FileSystemCollection | None:
        if op_parser.is_resource_op():
            return None
        folder_name = self._get_folder_name(op_parser)
        if folder_name is None:
            raise BadRequestError("Collection name must be specified")
        if folder_name in self._collection_cache:
            return self._collection_cache[folder_name]
        col = FileSystemCollection(self.store_path, folder_name)
        self._collection_cache[folder_name] = col
        return col

    def _validate(self, op_parser: StoreOperationParser):
        pass

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        self._validate(op_parser)
        collection = self._get_collection(op_parser)
        ncall, state = self._get_ncall(
            op_parser,
            collection,
            ResourceHelper(self.store_path),
        )
        if ncall is None:
            return super().run(operation=operation, context=context)
        nresult, nerror = ncall.invoke(return_error=True)
        result = self._convert_nresult(
            nresult, nerror, state, op_parser, collection
        )
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: StoreOperationParser,
        collection: FileSystemCollection | None,
        resource_helper: Any,
    ) -> tuple[NCall | None, dict | None]:
        call = None
        state = None
        nargs = op_parser.get_nargs()
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            args = {
                "folder": self._get_folder_name(op_parser),
                "config": get_collection_config(op_parser),
                "nargs": nargs,
            }
            call = NCall(resource_helper.create_collection, args)
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            args = {
                "folder": self._get_folder_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(resource_helper.drop_collection, args)
        # LIST COLLECTIONS
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            args = {"nargs": nargs}
            call = NCall(resource_helper.list_collections, args)
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            args = {"nargs": nargs}
            call = NCall(resource_helper.close, args)
        return call, state

    def _convert_nresult(
        self,
        nresult: Any,
        nerror: Any,
        state: dict | None,
        op_parser: StoreOperationParser,
        collection: FileSystemCollection | None,
    ) -> Any:
        result: Any = None
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            result = None
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            result = None
        # LIST COLLECTIONS
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            result = nresult
        return result


class ResourceHelper:
    store_path: str

    def __init__(self, store_path: str):
        self.store_path = store_path

    def create_collection(
        self, folder: str, config: ObjectCollectionConfig | None, nargs: Any
    ):
        folder_path = os.path.join(self.store_path, folder)
        os.makedirs(folder_path, exist_ok=True)

    def drop_collection(self, folder: str, nargs: Any):
        folder_path = os.path.join(self.store_path, folder)
        import shutil

        shutil.rmtree(folder_path)

    def list_collections(self, nargs) -> Any:
        nresult = []
        items = os.listdir(self.store_path)
        for item in items:
            if not os.path.isfile(item):
                nresult.append(item)
        return nresult

    def close(self, nargs: Any) -> Any:
        pass


class OperationConverter:
    processor: ItemProcessor

    def __init__(self, processor: ItemProcessor):
        self.processor = processor


class FileSystemCollection:
    converter: OperationConverter
    processor: ItemProcessor

    def __init__(self, store_path: str, collection_name: str):
        self.processor = ItemProcessor()
        self.converter = OperationConverter(self.processor)



"""
Secret Store on top of Document Store.
"""

from __future__ import annotations

__all__ = ["DocumentStoreProvider"]


import time
import uuid
from dataclasses import dataclass
from datetime import datetime
from typing import Any

from verse.core import Component, Context, NCall, Operation, Response
from verse.core.exceptions import (
    BadRequestError,
    NotFoundError,
    PreconditionFailedError,
)
from verse.internal.storage_core import (
    ItemProcessor,
    QueryProcessor,
    StoreFunction,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
)
from verse.ql import (
    And,
    Comparison,
    ComparisonOp,
    Field,
    OrderBy,
    OrderByTerm,
    Update,
)

from .._models import (
    SecretItem,
    SecretKey,
    SecretList,
    SecretProperties,
    SecretVersion,
)

SECRET = "secret"
VERSION = "version"


@dataclass
class SecretDocument:
    id: str
    pk: str
    tenant_id: str
    secret_id: str
    version: str
    metadata: dict | None
    value: str
    created_time: int
    type: str = SECRET

    def to_dict(self):
        return self.__dict__


@dataclass
class VersionDocument:
    id: str
    pk: str
    tenant_id: str
    secret_id: str
    version: str
    value: str
    created_time: int
    type: str = VERSION

    def to_dict(self):
        return self.__dict__


class DocumentStoreProvider(StoreProvider):
    store: Component
    tenant_id: str

    _init: bool
    _ainit: bool
    _processor: ItemProcessor
    _converter: OperationConverter

    def __init__(
        self,
        store: Component,
        tenant_id: str = "default",
        **kwargs,
    ):
        """Initialize.

        Args:
            store:
                Document Store component.
            tenant_id:
                Tenant id for multi-tenancy, defaults to "default".
        """
        self.store = store
        self.tenant_id = tenant_id

        self._init = False
        self._ainit = False
        self._processor = ItemProcessor()
        self._converter = OperationConverter(self.tenant_id)

    def init(self, context: Context | None = None) -> None:
        if self._init:
            return
        self.store.init(context=context)
        self._init = True

    async def ainit(self, context: Context | None = None) -> None:
        if self._ainit:
            return
        await self.store.ainit(context=context)
        self._ainit = True

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        ncall, state = self._get_operation(
            op_parser,
            self._converter,
            ClientHelper(self.store, self._converter),
        )
        if ncall is None:
            return super().run(operation)
        if isinstance(ncall, NCall):
            operation = ncall.invoke()
        else:
            operation = ncall
        nresult = self.store.run(operation=operation, context=context)
        result = self._convert_nresult(
            nresult, state, op_parser, self._processor
        )
        return Response(result=result, native=dict(result=nresult, call=ncall))

    async def arun(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        await self.ainit(context=context)
        op_parser = self.get_op_parser(operation)
        ncall, state = self._get_operation(
            op_parser,
            self._converter,
            AsyncClientHelper(self.store, self._converter),
        )
        if ncall is None:
            return await super().arun(operation)
        if isinstance(ncall, NCall):
            operation = await ncall.ainvoke()
        else:
            operation = ncall
        nresult = await self.store.arun(operation=operation, context=context)
        result = self._convert_nresult(
            nresult, state, op_parser, self._processor
        )
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_operation(
        self,
        op_parser: StoreOperationParser,
        converter: OperationConverter,
        helper: Any,
    ) -> tuple[Operation | NCall | None, dict | None]:
        operation: NCall | Operation | None = None
        state: dict | None = None
        # GET value
        if (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_value_attribute()
        ):
            id = op_parser.get_id_as_str()
            version = op_parser.get_version()
            operation = converter.convert_get(id, version)
        # GET metadata
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_metadata_attribute()
        ):
            id = op_parser.get_id_as_str()
            operation = converter.convert_get_metadata(id)
        # GET versions
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_versions_attribute()
        ):
            id = op_parser.get_id_as_str()
            operation = converter.convert_get_versions(id)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            id = op_parser.get_id_as_str()
            value = op_parser.get_value()
            metadata: dict | None = op_parser.get_metadata()
            metadata = {} if metadata is None else metadata
            exists = op_parser.get_where_exists()
            version = str(uuid.uuid4())
            created_time = int(time.time())
            args: dict[str, Any] = {
                "id": id,
                "value": value,
                "metadata": metadata,
                "exists": exists,
                "version": version,
                "created_time": created_time,
            }
            operation = NCall(helper.put, args)
            state = {"version": version}
        # UPDATE value
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_value_update()
        ):
            id = op_parser.get_id_as_str()
            value = op_parser.get_updated_value()
            version = str(uuid.uuid4())
            created_time = int(time.time())
            args = {
                "id": id,
                "value": value,
                "version": version,
                "created_time": created_time,
            }
            operation = NCall(helper.update_value, args)
            state = {"version": version}
        # UPDATE metadata
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_metadata_update()
        ):
            id = op_parser.get_id_as_str()
            metadata = op_parser.get_updated_metadata()
            metadata = {} if metadata is None else metadata
            operation = converter.convert_update_metadata(id, metadata)
            state = {"metadata": metadata}
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            id = op_parser.get_id_as_str()
            args = {"id": id}
            operation = NCall(helper.delete, args)
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            operation = converter.convert_query()
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            operation = converter.convert_query()
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            operation = StoreOperation.close()
        return operation, state

    def _convert_nresult(
        self,
        nresult: Any,
        state: dict | None,
        op_parser: StoreOperationParser,
        processor: ItemProcessor,
    ) -> Any:
        result: Any = None
        nresult = nresult.result
        # GET value
        if (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_value_attribute()
        ):
            result = SecretItem(
                key=SecretKey(
                    id=nresult.value["secret_id"],
                    version=nresult.value["version"],
                ),
                value=nresult.value["value"],
            )
        # GET metadata
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_metadata_attribute()
        ):
            result = SecretItem(
                key=SecretKey(id=nresult.value["secret_id"]),
                metadata=nresult.value["metadata"],
                properties=SecretProperties(
                    created_time=Helper.convert_to_datetime(
                        nresult.value["created_time"]
                    )
                ),
            )
        # GET versions
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_versions_attribute()
        ):
            versions: list = []
            for item in nresult.items:
                id = item.value["secret_id"]
                versions.append(
                    SecretVersion(
                        version=item.value["version"],
                        created_time=Helper.convert_to_datetime(
                            item.value["created_time"]
                        ),
                    )
                )
            if len(versions) == 0:
                raise NotFoundError
            versions = sorted(
                versions,
                key=lambda s: s.created_time,
                reverse=True,
            )
            return SecretItem(key=SecretKey(id=id), versions=versions)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            result = SecretItem(
                key=SecretKey(
                    id=op_parser.get_id_as_str(),
                    version=state["version"] if state is not None else None,
                )
            )
        # UPDATE value
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_value_update()
        ):
            result = SecretItem(
                key=SecretKey(
                    id=op_parser.get_id_as_str(),
                    version=state["version"] if state is not None else None,
                )
            )
        # UPDATE metadata
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_metadata_update()
        ):
            result = SecretItem(
                key=SecretKey(id=op_parser.get_id_as_str()),
                metadata=state["metadata"] if state is not None else None,
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            result = None
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            items = []
            for item in nresult.items:
                items.append(
                    SecretItem(
                        key=SecretKey(id=item.value["secret_id"]),
                        metadata=item.value["metadata"],
                        properties=SecretProperties(
                            created_time=Helper.convert_to_datetime(
                                item.value["created_time"]
                            )
                        ),
                    )
                )
            order_by = (
                op_parser.get_order_by()
                if op_parser.get_order_by() is not None
                else OrderBy(terms=[OrderByTerm(field="$id")])
            )
            items = QueryProcessor.query_items(
                items=items,
                select=op_parser.get_select(),
                where=op_parser.get_where(),
                order_by=order_by,
                limit=op_parser.get_limit(),
                offset=op_parser.get_offset(),
                field_resolver=processor.resolve_root_field,
            )
            result = SecretList(items=items)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            items = []
            for item in nresult.items:
                items.append(
                    SecretItem(
                        key=SecretKey(id=item.value["secret_id"]),
                        metadata=item.value["metadata"],
                        properties=SecretProperties(
                            created_time=Helper.convert_to_datetime(
                                item.value["created_time"]
                            )
                        ),
                    )
                )
            result = QueryProcessor.count_items(
                items=items,
                where=op_parser.get_where(),
                field_resolver=processor.resolve_root_field,
            )
        return result


class Helper:
    @staticmethod
    def convert_to_datetime(timestamp: int) -> datetime:
        return datetime.fromtimestamp(timestamp).astimezone()


class ClientHelper:
    store: Component
    converter: OperationConverter

    def __init__(
        self,
        store: Component,
        converter: OperationConverter,
    ):
        self.store = store
        self.converter = converter

    def put(
        self,
        id: str,
        value: str,
        metadata: dict | None,
        exists: bool | None,
        version: str,
        created_time: int,
    ) -> Operation:
        get_result = None
        try:
            get_response = self.store.run(
                StoreOperation.get(key=self.converter.convert_item_key(id))
            )
            get_result = get_response.result
            if exists is False:
                raise PreconditionFailedError
        except NotFoundError:
            pass
        if exists is True and get_result is None:
            raise PreconditionFailedError

        return self.converter.convert_put(
            id, value, version, metadata, created_time, get_result
        )

    def update_value(
        self,
        id: str,
        value: str,
        version: str,
        created_time: int,
    ) -> Operation:
        get_response = self.store.run(
            StoreOperation.get(key=self.converter.convert_item_key(id))
        )
        get_result = get_response.result
        return self.converter.convert_update_value(
            id, value, version, created_time, get_result
        )

    def delete(self, id: str) -> Operation:
        response = self.store.run(self.converter.convert_query_secret(id))
        if response is not None and response.result is not None:
            if len(response.result.items) == 0:
                raise NotFoundError
            return self.converter.convert_delete(response.result.items)
        raise BadRequestError("Unexcepted response")


class AsyncClientHelper:
    store: Component
    converter: OperationConverter

    def __init__(
        self,
        store: Component,
        converter: OperationConverter,
    ):
        self.store = store
        self.converter = converter

    async def put(
        self,
        id: str,
        value: str,
        metadata: dict | None,
        exists: bool | None,
        version: str,
        created_time: int,
    ) -> Operation:
        get_result = None
        try:
            get_response = await self.store.arun(
                StoreOperation.get(key=self.converter.convert_item_key(id))
            )
            get_result = get_response.result
            if exists is False:
                raise PreconditionFailedError
        except NotFoundError:
            pass
        if exists is True and get_result is None:
            raise PreconditionFailedError

        return self.converter.convert_put(
            id, value, version, metadata, created_time, get_result
        )

    async def update_value(
        self,
        id: str,
        value: str,
        version: str,
        created_time: int,
    ) -> Operation:
        get_response = await self.store.arun(
            StoreOperation.get(key=self.converter.convert_item_key(id))
        )
        get_result = get_response.result
        return self.converter.convert_update_value(
            id, value, version, created_time, get_result
        )

    async def delete(self, id: str) -> Operation:
        response = await self.store.arun(
            self.converter.convert_query_secret(id)
        )
        if response is not None and response.result is not None:
            if len(response.result.items) == 0:
                raise NotFoundError
            return self.converter.convert_delete(response.result.items)
        raise BadRequestError("Unexcepted response")


class OperationConverter:
    tenant_id: str

    def __init__(self, tenant_id: str, **kwargs: Any):
        self.tenant_id = tenant_id

    def convert_secret_id(self, id: str):
        return f"{SECRET}-{self.tenant_id}-{id}"

    def convert_version_id(self, id: str, version: str | None = None):
        vid = f"{VERSION}-{self.tenant_id}-{id}"
        if version is not None:
            vid = f"{vid}-{version}"
        return vid

    def convert_pk(self):
        return self.tenant_id

    def convert_item_key(self, id: str):
        key = dict()
        key["$id"] = self.convert_secret_id(id)
        key["$pk"] = self.convert_pk()
        return key

    def convert_version_key(self, id: str, version: str | None = None):
        key = dict()
        key["$id"] = self.convert_version_id(id, version)
        key["$pk"] = self.convert_pk()
        return key

    def convert_query(self) -> Operation:
        return StoreOperation.query(
            where=And(
                lexpr=Comparison(
                    lexpr=Field(path="$pk"),
                    op=ComparisonOp.EQ,
                    rexpr=self.tenant_id,
                ),
                rexpr=Comparison(
                    lexpr=Field(path="type"),
                    op=ComparisonOp.EQ,
                    rexpr=SECRET,
                ),
            )
        )

    def convert_query_secret(self, id: str) -> Operation:
        return StoreOperation.query(
            where=And(
                lexpr=Comparison(
                    lexpr=Field(path="$pk"),
                    op=ComparisonOp.EQ,
                    rexpr=self.tenant_id,
                ),
                rexpr=Comparison(
                    lexpr=Field(path="secret_id"),
                    op=ComparisonOp.EQ,
                    rexpr=id,
                ),
            )
        )

    def convert_get(self, id: str, version: str | None) -> Operation:
        if version is None:
            return StoreOperation.get(key=self.convert_item_key(id))
        else:
            return StoreOperation.get(
                key=self.convert_version_key(id, version)
            )

    def convert_get_metadata(self, id: str) -> Operation:
        return StoreOperation.get(key=self.convert_item_key(id))

    def convert_get_versions(self, id: str) -> Operation:
        return StoreOperation.query(
            where=And(
                lexpr=And(
                    lexpr=Comparison(
                        lexpr=Field(path="$pk"),
                        op=ComparisonOp.EQ,
                        rexpr=self.tenant_id,
                    ),
                    rexpr=Comparison(
                        lexpr=Field(path="type"),
                        op=ComparisonOp.EQ,
                        rexpr=VERSION,
                    ),
                ),
                rexpr=Comparison(
                    lexpr=Field(path="secret_id"),
                    op=ComparisonOp.EQ,
                    rexpr=id,
                ),
            )
        )

    def convert_put(
        self,
        id: str,
        value: str,
        version: str,
        metadata: dict | None,
        created_time: int,
        get_result: Any,
    ) -> Operation:
        operations = []
        if get_result is not None:
            if get_result.properties.etag is not None:
                where = Comparison(
                    lexpr=Field(path="$etag"),
                    op=ComparisonOp.EQ,
                    rexpr=get_result.properties.etag,
                )
            else:
                where = Comparison(
                    lexpr=Field(path="version"),
                    op=ComparisonOp.EQ,
                    rexpr=get_result.value["version"],
                )
            update = Update()
            update.put("metadata", metadata)
            update.put("version", version)
            update.put("value", value)
            operations.append(
                StoreOperation.update(
                    key=self.convert_item_key(id), set=update, where=where
                )
            )
        else:
            operations.append(
                StoreOperation.put(
                    value=SecretDocument(
                        id=self.convert_secret_id(id),
                        pk=self.convert_pk(),
                        tenant_id=self.tenant_id,
                        secret_id=id,
                        version=version,
                        metadata=metadata,
                        value=value,
                        created_time=created_time,
                    ).to_dict(),
                    where=StoreFunction.not_exists(),
                )
            )
        operations.append(
            StoreOperation.put(
                value=VersionDocument(
                    id=self.convert_version_id(id, version),
                    pk=self.convert_pk(),
                    tenant_id=self.tenant_id,
                    secret_id=id,
                    version=version,
                    value=value,
                    created_time=created_time,
                ).to_dict()
            )
        )
        return StoreOperation.transact(operations)

    def convert_update_value(
        self,
        id: str,
        value: str,
        version: str,
        created_time: int,
        get_result: Any,
    ) -> Operation:
        operations = []
        if get_result.properties.etag is not None:
            where = Comparison(
                lexpr=Field(path="$etag"),
                op=ComparisonOp.EQ,
                rexpr=get_result.properties.etag,
            )
        else:
            where = Comparison(
                lexpr=Field(path="version"),
                op=ComparisonOp.EQ,
                rexpr=get_result.value["version"],
            )
        update = Update()
        update.put("version", version)
        update.put("value", value)
        operations.append(
            StoreOperation.update(
                key=self.convert_item_key(id), set=update, where=where
            )
        )
        operations.append(
            StoreOperation.put(
                value=VersionDocument(
                    id=self.convert_version_id(id, version),
                    pk=self.convert_pk(),
                    tenant_id=self.tenant_id,
                    secret_id=id,
                    version=version,
                    value=value,
                    created_time=created_time,
                ).to_dict()
            )
        )
        return StoreOperation.transact(operations)

    def convert_update_metadata(
        self, id: str, metadata: dict | None
    ) -> Operation:
        update = Update()
        update.put("metadata", metadata)
        return StoreOperation.update(key=self.convert_item_key(id), set=update)

    def convert_delete(self, items: list[Any]):
        operations = []
        for item in items:
            if item.value["type"] == SECRET:
                if item.properties.etag is not None:
                    where = Comparison(
                        lexpr=Field(path="$etag"),
                        op=ComparisonOp.EQ,
                        rexpr=item.properties.etag,
                    )
                else:
                    where = Comparison(
                        lexpr=Field(path="version"),
                        op=ComparisonOp.EQ,
                        rexpr=item.value["version"],
                    )
                operations.append(
                    StoreOperation.delete(key=item.key, where=where)
                )
            else:
                operations.append(StoreOperation.delete(key=item.key))
        return StoreOperation.transact(operations)



from enum import Enum
from typing import Any, Union

from verse.core import DataModel
from verse.internal.storage_core import Index

DocumentKeyType = Union[str, int, float, bool]


class DocumentType(str, Enum):
    """Document type."""

    JSON = "json"
    TEXT = "text"
    BYTES = "bytes"


class DocumentFieldType(str, Enum):
    """Document field type."""

    STRING = "string"
    NUMBER = "number"
    BOOLEAN = "boolean"
    OBJECT = "object"
    ARRAY = "array"
    NULL = "null"


class DocumentKey(DataModel):
    """Document key."""

    id: DocumentKeyType
    """Document id."""

    pk: DocumentKeyType | None = None
    """Partition key."""


class DocumentProperties(DataModel):
    """Document properties."""

    etag: str | None = None
    """Document ETag."""


class DocumentItem(DataModel):
    """Document item."""

    key: DocumentKey
    """Document key."""

    value: dict[str, Any] | None = None
    """Document value."""

    properties: DocumentProperties | None = None
    """Document properties."""


class DocumentList(DataModel):
    """Document list."""

    items: list[DocumentItem]
    """List of documents."""


class DocumentCollectionConfig(DataModel):
    """Document collection config."""

    id_field: str | None = None
    """Id field name."""

    id_type: DocumentFieldType | None = None
    """Id field type."""

    pk_field: str | None = None
    """Partition key field name."""

    pk_type: DocumentFieldType | None = None
    """Partition key field type."""

    value_field: str | None = None
    """Value field name."""

    value_type: DocumentType = DocumentType.JSON
    """Value type."""

    indexes: list[Index] | None = None
    """Indexes."""

    nconfig: dict | None = None
    """Native config parameters."""


class DocumentQueryConfig(DataModel):
    """Query config."""

    paging: bool | None = False
    """A value indicating whether the results should be paged."""

    page_size: int | None = None
    """Page size."""



from __future__ import annotations

from typing import Any

from verse.core import DataModel, Operation, Response
from verse.internal.storage_core import Index, StoreComponent, StoreOperation
from verse.ql import Expression, OrderBy, Select, Update

from ._models import (
    DocumentCollectionConfig,
    DocumentItem,
    DocumentKey,
    DocumentKeyType,
    DocumentList,
)


class DocumentStore(StoreComponent):
    collection: str | None
    id_map_field: str | dict | None
    pk_map_field: str | dict | None

    def __init__(
        self,
        collection: str | None = None,
        id_map_field: str | dict | None = "id",
        pk_map_field: str | dict | None = "pk",
        **kwargs,
    ):
        """Initialize.

        Args:
            collection:
                Default collection name.
            id_map_field:
                Field in the document to map into id.
                To specify for multiple collections, use a dictionary
                where the key is the collection name and the value
                is the field.
            pk_map_field:
                Field in the document to map into pk.
                To specify for multiple collections, use a dictionary
                where the key is the collection name and the value
                is the field.
        """
        self.collection = collection
        self.id_map_field = id_map_field
        self.pk_map_field = pk_map_field

        super().__init__(**kwargs)

    def _attach_component_parameters(self):
        setattr(self.provider, "collection", self.collection)
        setattr(self.provider, "id_map_field", self.id_map_field)
        setattr(self.provider, "pk_map_field", self.pk_map_field)

    def create_collection(
        self,
        collection: str | None = None,
        config: dict | DocumentCollectionConfig | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Create collection.

        Args:
            collection:
                Collection name.
            config:
                Collection config.
        """
        return self._run_internal(StoreOperation.CREATE_COLLECTION, locals())

    def drop_collection(
        self,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Drop collection.

        Args:
            collection: Collection name.
        """
        return self._run_internal(StoreOperation.DROP_COLLECTION, locals())

    def list_collections(
        self,
        **kwargs: Any,
    ) -> Response[list[str]]:
        """List collections.

        Returns:
            List of collection names.
        """
        return self._run_internal(StoreOperation.LIST_COLLECTIONS, locals())

    def create_index(
        self,
        index: dict | Index,
        collection: str | None = None,
    ) -> Response[None]:
        """Create index.

        Args:
            index:
                Index to create.
            collection:
                Collection name.
        """
        return self._run_internal(StoreOperation.CREATE_INDEX, locals())

    def drop_index(
        self,
        index: dict | Index | None = None,
        collection: str | None = None,
    ) -> Response[None]:
        """Drop index.

        Args:
            index:
                Index to drop.
            collection:
                Collection name.
        """
        return self._run_internal(StoreOperation.DROP_INDEX, locals())

    def list_indexes(
        self,
        collection: str | None = None,
    ) -> Response[list[Index]]:
        """List indexes.

        Args:
            collection:
                Collection name.
        """
        return self._run_internal(StoreOperation.LIST_INDEXES, locals())

    def get(
        self,
        key: DocumentKeyType | dict | DocumentKey,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[DocumentItem]:
        """Get document.

        Args:
            key:
                Document key.
            collection:
                Collection name.

        Returns:
            Document item.

        Raises:
            NotFoundError:
                Document not found.
        """
        return self._run_internal(StoreOperation.GET, locals())

    def put(
        self,
        value: dict[str, Any] | DataModel,
        key: DocumentKeyType | dict | DocumentKey | None = None,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[DocumentItem]:
        """Put document.

        Args:
            value:
                Document.
            key:
                Document key.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            Document item.

        Raises:
            PreconditionFailedError:
                Condition failed.
        """
        return self._run_internal(StoreOperation.PUT, locals())

    def update(
        self,
        key: DocumentKeyType | dict | DocumentKey,
        set: str | Update,
        where: str | Expression | None = None,
        returning: bool | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[DocumentItem]:
        """Update document.

        Args:
            key:
                Document key.
            set:
                Update expression.
            where:
                Condition expression.
            returning:
                A value indicating whether updated document is returned.
            collection:
                Collection name.

        Returns:
            Document item.

        Raises:
            NotFoundError:
                Document not found.
            PreconditionFailedError:
                Condition failed.
        """
        return self._run_internal(StoreOperation.UPDATE, locals())

    def delete(
        self,
        key: DocumentKeyType | dict | DocumentKey,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Delete document.

        Args:
            key:
                Document key.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            None.

        Raises:
            NotFoundError:
                Document not found.
            PreconditionFailedError:
                Condition failed.
        """
        return self._run_internal(StoreOperation.DELETE, locals())

    def query(
        self,
        select: str | Select | None = None,
        where: str | Expression | None = None,
        order_by: str | OrderBy | None = None,
        limit: int | None = None,
        offset: int | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[DocumentList]:
        """Query documents.

        Args:
            select:
                Select expression.
            where:
                Condition expression.
            order_by:
                Order by expression.
            limit:
                Query limit.
            offset:
                Query offset.
            collection:
                Collection name.

        Returns:
            Document list with items.
        """
        return self._run_internal(StoreOperation.QUERY, locals())

    def count(
        self,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[int]:
        """Count documents.

        Args:
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            Count of documents.
        """
        return self._run_internal(StoreOperation.COUNT, locals())

    def batch(
        self,
        operations: list[Operation],
        **kwargs,
    ) -> Response[list[Any]]:
        """Execute batch.

        Args:
            operations:
                List of document operations.

        Returns:
            Batch operation results.
        """
        return self._run_internal(StoreOperation.BATCH, locals())

    def transact(
        self,
        operations: list[Operation],
        **kwargs,
    ) -> Response[list[Any]]:
        """Execute transaction.

        Args:
            operations:
                List of document operations.

        Returns:
            Transaction results.

        Raises:
            ConflictError:
                Transaction failed.
        """
        return self._run_internal(StoreOperation.TRANSACT, locals())

    def close(
        self,
        **kwargs: Any,
    ) -> Response[None]:
        """Close client.

        Returns:
            None.
        """
        return self._run_internal(StoreOperation.CLOSE, locals())

    async def acreate_collection(
        self,
        collection: str | None = None,
        config: dict | DocumentCollectionConfig | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Create collection.

        Args:
            collection:
                Collection name.
            config:
                Collection config.

        Returns:
            None.
        """
        return await self._arun_internal(
            StoreOperation.CREATE_COLLECTION, locals()
        )

    async def adrop_collection(
        self,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Drop collection.

        Args:
            collection: Collection name.

        Returns:
            None.
        """
        return await self._arun_internal(
            StoreOperation.DROP_COLLECTION, locals()
        )

    async def alist_collections(
        self,
        **kwargs: Any,
    ) -> Response[list[str]]:
        """List collections.

        Returns:
            List of collection names.
        """
        return await self._arun_internal(
            StoreOperation.LIST_COLLECTIONS, locals()
        )

    async def acreate_index(
        self,
        index: dict | Index,
        collection: str | None = None,
    ) -> Response[None]:
        """Create index.

        Args:
            index:
                Index to create.
            collection:
                Collection name.
        """
        return await self._arun_internal(StoreOperation.CREATE_INDEX, locals())

    async def adrop_index(
        self,
        index: dict | Index | None = None,
        collection: str | None = None,
    ) -> Response[None]:
        """Drop index.

        Args:
            index:
                Index to drop.
            collection:
                Collection name.
        """
        return await self._arun_internal(StoreOperation.DROP_INDEX, locals())

    async def alist_indexes(
        self,
        collection: str | None = None,
    ) -> Response[list[Index]]:
        """List indexes.

        Args:
            collection:
                Collection name.
        """
        return await self._arun_internal(StoreOperation.LIST_INDEXES, locals())

    async def aget(
        self,
        key: DocumentKeyType | dict | DocumentKey,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[DocumentItem]:
        """Get document.

        Args:
            key:
                Document key.
            collection:
                Collection name.

        Returns:
            Document item.

        Raises:
            NotFoundError:
                Document not found.
        """
        return await self._arun_internal(StoreOperation.GET, locals())

    async def aput(
        self,
        value: dict[str, Any] | DataModel,
        key: DocumentKeyType | dict | DocumentKey | None = None,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[DocumentItem]:
        """Put document.

        Args:
            value:
                Document.
            key:
                Document key.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            Document item.

        Raises:
            PreconditionFailedError:
                Condition failed.
        """
        return await self._arun_internal(StoreOperation.PUT, locals())

    async def aupdate(
        self,
        key: DocumentKeyType | dict | DocumentKey,
        set: str | Update,
        where: str | Expression | None = None,
        returning: bool | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[DocumentItem]:
        """Update document.

        Args:
            key:
                Document key.
            set:
                Update expression.
            where:
                Condition expression.
            returning:
                A value indicating whether updated document is returned.
            collection:
                Collection name.

        Returns:
            Document item.

        Raises:
            NotFoundError:
                Document not found.
            PreconditionFailedError:
                Condition failed.
        """
        return await self._arun_internal(StoreOperation.UPDATE, locals())

    async def adelete(
        self,
        key: DocumentKeyType | dict | DocumentKey,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Delete document.

        Args:
            key:
                Document key.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            None.

        Raises:
            NotFoundError:
                Document not found.
            PreconditionFailedError:
                Condition failed.
        """
        return await self._arun_internal(StoreOperation.DELETE, locals())

    async def aquery(
        self,
        select: str | Select | None = None,
        where: str | Expression | None = None,
        order_by: str | OrderBy | None = None,
        limit: int | None = None,
        offset: int | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[DocumentList]:
        """Query documents.

        Args:
            select:
                Select expression.
            where:
                Condition expression.
            order_by:
                Order by expression.
            limit:
                Query limit.
            offset:
                Query offset.
            collection:
                Collection name.

        Returns:
            Document list with items.
        """
        return await self._arun_internal(StoreOperation.QUERY, locals())

    async def acount(
        self,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[int]:
        """Count documents.

        Args:
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            Count of documents.
        """
        return await self._arun_internal(StoreOperation.COUNT, locals())

    async def abatch(
        self,
        operations: list[Operation],
        **kwargs: Any,
    ) -> Response[list[Any]]:
        """Execute batch.

        Args:
            operations:
                List of document operations.

        Returns:
            Batch operation results.
        """
        return await self._arun_internal(StoreOperation.BATCH, locals())

    async def atransact(
        self,
        operations: list[Operation],
        **kwargs: Any,
    ) -> Response[list[Any]]:
        """Execute transaction.

        Args:
            operations:
                List of document operations.

        Returns:
            Transaction results.

        Raises:
            ConflictError:
                Transaction failed.
        """
        return await self._arun_internal(StoreOperation.TRANSACT, locals())

    async def aclose(
        self,
        **kwargs: Any,
    ) -> Response[None]:
        """Close async client.

        Returns:
            None.
        """
        return await self._arun_internal(StoreOperation.CLOSE, locals())


class DocumentOperation:
    @staticmethod
    def get(
        key: DocumentKeyType | dict | DocumentKey,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Operation:
        """Get operation.

        Args:
            key:
                Document key.
            collection:
                Collection name.

        Returns:
            Get operation.
        """
        return StoreOperation.get(
            key=key,
            collection=collection,
            **kwargs,
        )

    @staticmethod
    def put(
        value: dict[str, Any] | DataModel,
        key: DocumentKeyType | dict | DocumentKey | None = None,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Operation:
        """Put operation.

        Args:
            value:
                Document value.
            key:
                Document key.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            _description_
        """
        return StoreOperation.put(
            key=key,
            value=value,
            where=where,
            collection=collection,
            **kwargs,
        )

    @staticmethod
    def update(
        key: DocumentKeyType | dict | DocumentKey,
        set: str | Update,
        where: str | Expression | None = None,
        returning: bool | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Operation:
        """Update operation.

        Args:
            key:
                Document key.
            set:
                Update expression.
            where:
                Condition expression.
            returning:
                A value indicating whether the updated document is returned.
            collection:
                Collection name.

        Returns:
            Update operation.
        """
        return StoreOperation.update(
            key=key,
            set=set,
            where=where,
            returning=returning,
            collection=collection,
            **kwargs,
        )

    @staticmethod
    def delete(
        key: DocumentKeyType | dict | DocumentKey,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Operation:
        """Delete operation.

        Args:
            key:
                Document key.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            Delete operation.
        """
        return StoreOperation.delete(
            key=key,
            where=where,
            collection=collection,
            **kwargs,
        )


class DocumentBatch(DataModel):
    operations: list[Operation] = []

    def get(
        self,
        key: DocumentKeyType | dict | DocumentKey,
        collection: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Get operation.

        Args:
            key:
                Document key.
            collection:
                Collection name.

        Returns:
            Get operation.
        """
        self.operations.append(
            Operation.normalize(name=StoreOperation.GET, args=locals())
        )

    def put(
        self,
        value: dict[str, Any] | DataModel,
        key: DocumentKeyType | dict | DocumentKey | None = None,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Put operation.

        Args:
            value:
                Document value.
            key:
                Document key.
            where:
                Condition expression.
            collection:
                Collection name.
        """
        self.operations.append(
            Operation.normalize(name=StoreOperation.PUT, args=locals())
        )

    def update(
        self,
        key: DocumentKeyType | dict | DocumentKey,
        set: str | Update,
        where: str | Expression | None = None,
        returning: bool | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Update operation.

        Args:
            key:
                Document key.
            set:
                Update expression.
            where:
                Condition expression.
            returning:
                A value indicating whether the updated document is returned.
            collection:
                Collection name.

        Returns:
            Update operation.
        """
        self.operations.append(
            Operation.normalize(name=StoreOperation.UPDATE, args=locals())
        )

    def delete(
        self,
        key: DocumentKeyType | dict | DocumentKey,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Delete operation.

        Args:
            key:
                Document key.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            Delete operation.
        """
        self.operations.append(
            Operation.normalize(name=StoreOperation.PUT, args=locals())
        )


class DocumentTransaction(DataModel):
    operations: list[Operation] = []

    def get(
        self,
        key: DocumentKeyType | dict | DocumentKey,
        collection: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Get operation.

        Args:
            key:
                Document key.
            collection:
                Collection name.

        Returns:
            Get operation.
        """
        self.operations.append(
            Operation.normalize(name=StoreOperation.GET, args=locals())
        )

    def put(
        self,
        value: dict[str, Any] | DataModel,
        key: DocumentKeyType | dict | DocumentKey | None = None,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Put operation.

        Args:
            value:
                Document value.
            key:
                Document key.
            where:
                Condition expression.
            collection:
                Collection name.
        """
        self.operations.append(
            Operation.normalize(name=StoreOperation.PUT, args=locals())
        )

    def update(
        self,
        key: DocumentKeyType | dict | DocumentKey,
        set: str | Update,
        where: str | Expression | None = None,
        returning: bool | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Update operation.

        Args:
            key:
                Document key.
            set:
                Update expression.
            where:
                Condition expression.
            returning:
                A value indicating whether the updated document is returned.
            collection:
                Collection name.

        Returns:
            Update operation.
        """
        self.operations.append(
            Operation.normalize(name=StoreOperation.UPDATE, args=locals())
        )

    def delete(
        self,
        key: DocumentKeyType | dict | DocumentKey,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Delete operation.

        Args:
            key:
                Document key.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            Delete operation.
        """
        self.operations.append(
            Operation.normalize(name=StoreOperation.PUT, args=locals())
        )



from datetime import datetime
from enum import Enum
from typing import Any

from verse.core import DataModel


class ObjectStoreClass(str, Enum):
    """Object Store class."""

    HOT = "hot"
    COOL = "cool"
    COLD = "cold"
    ARCHIVE = "archive"


class ObjectKey(DataModel):
    """Object key."""

    id: str
    """Object id."""

    version: str | None = None
    """Object version."""


class ObjectProperties(DataModel):
    """Object properties."""

    cache_control: str | None = None
    """Cache control."""

    content_disposition: str | None = None
    """Content disposition."""

    content_encoding: str | None = None
    """Content encoding."""

    content_language: str | None = None
    """Content language."""

    content_length: int | None = None
    """Content length."""

    content_md5: str | None = None
    """Content md5."""

    content_type: str | None = None
    """Content type."""

    crc32c: str | None = None
    """Crc32c value."""

    expires: datetime | None = None
    """Expiration time."""

    last_modified: datetime | None = None
    """Last modified time."""

    etag: str | None = None
    """Etag value."""

    storage_class: str | None = None
    """Storage class."""


class ObjectVersion(DataModel):
    """Object version."""

    version: str | None
    """Version id."""

    properties: ObjectProperties | None = None
    """Object properties."""

    metadata: dict | None = None
    """Custom metadata."""

    latest: bool | None = None
    """A value indicating whether this is the latest version."""


class ObjectItem(DataModel):
    """Object item."""

    key: ObjectKey
    """Object key."""

    value: bytes | None = None
    """Object value."""

    metadata: dict | None = None
    """Object metadata."""

    properties: ObjectProperties | None = None
    """Object properties."""

    versions: list[ObjectVersion] | None = None
    """Object vesions."""

    url: str | None = None
    """Object url."""


class ObjectList(DataModel):
    """Object list."""

    items: list[ObjectItem]
    """List of items."""

    continuation: str | None = None
    """Continuation token."""

    prefixes: list[str] | None = None
    """List of prefixes."""


class ObjectCollectionConfig(DataModel):
    """Object collection config."""

    acl: str | None = None
    """Access control."""

    versioned: bool | None = None
    """A value indicating whether the store is versioned."""

    nconfig: dict[str, Any] | None = None
    """Native config for providers."""


class ObjectTransferConfig(DataModel):
    """Transfer config."""

    multipart: bool | None = False
    """A value indicating whether the transfer should be multi-part."""

    chunksize: int | None = None
    """Chunk size for multi-part transfer."""

    concurrency: int | None = None
    """Number of concurrent transfer."""

    nconfig: dict[str, Any] | None = None
    """Native config for providers."""


class ObjectSource(DataModel):
    """Object source."""

    id: str
    """Object id."""

    version: str | None = None
    """Object version."""

    collection: str | None = None
    """Collection name."""


class ObjectQueryConfig(DataModel):
    """Query config."""

    paging: bool | None = False
    """A value indicating whether the results should be paged."""

    page_size: int | None = None
    """Page size."""



"""
Document Store on Google Firestore.
"""

from __future__ import annotations

__all__ = ["GoogleFirestore"]

from typing import Any

from google.api_core.exceptions import AlreadyExists, NotFound
from google.cloud import firestore
from google.cloud.firestore import (
    ExistsOption,
    async_transactional,
    transactional,
)
from google.cloud.firestore_v1.base_query import And as AndFilter
from google.cloud.firestore_v1.base_query import BaseFilter, FieldFilter
from google.cloud.firestore_v1.base_query import Or as OrFilter

from verse.core import Context, NCall, Operation, Response
from verse.core.exceptions import (
    BadRequestError,
    ConflictError,
    NotFoundError,
    PreconditionFailedError,
)
from verse.internal.storage_core import (
    ItemProcessor,
    ParameterParser,
    QueryProcessor,
    StoreFunctionName,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
    Validator,
)
from verse.ql import (
    And,
    Comparison,
    ComparisonOp,
    Expression,
    Field,
    Function,
    FunctionNamespace,
    Or,
    OrderBy,
    OrderByDirection,
    Select,
    Update,
    UpdateOp,
    Value,
)

from .._helper import (
    build_item_from_parts,
    build_item_from_value,
    build_query_result,
    get_collection_config,
)
from .._models import DocumentCollectionConfig


class GoogleFirestore(StoreProvider):
    collection: str | None = None
    project: str
    database: str | None
    service_account_info: str | None = None
    service_account_file: str | None = None
    access_token: str | None = None
    id_map_field: str | dict | None
    pk_map_field: str | dict | None
    etag_embed_field: str | dict | None
    suppress_fields: list[str] | None = None
    nparams: dict[str, Any] = dict()

    _credential: Any
    _db_client: Any
    _adb_client: Any
    _collection_cache: dict[str, FirestoreCollection]
    _acollection_cache: dict[str, FirestoreCollection]

    def __init__(
        self,
        project: str,
        database: str | None = "(default)",
        collection: str | None = None,
        service_account_info: str | None = None,
        service_account_file: str | None = None,
        access_token: str | None = None,
        id_map_field: str | dict | None = "id",
        pk_map_field: str | dict | None = "pk",
        etag_embed_field: str | dict | None = "_etag",
        suppress_fields: list[str] | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            project:
                Google project name.
            database:
                Firestore database name.
            collection:
                Firestore collection name which
                maps to document store collection.
            service_account_info:
                Google service account info with serialized credentials.
            service_account_file:
                Google service account file with credentials.
            access_token:
                Google access token.
            id_map_field:
                Field in the document to map into id.
                To specify for multiple collections, use a dictionary
                where the key is the collection name and the value
                is the field.
            pk_map_field:
                Field in the document to map into pk.
                To specify for multiple collections, use a dictionary
                where the key is the collection name and the value
                is the field.
            etag_embed_field:
                Field to store the generated ETAG value.
                To specify for multiple collections, use a dictionary
                where the key is the collection name and the value
                is the field.
            suppress_fields:
                List of fields to supress when results are returned.
            nparams:
                Native params to Google Firestore client.
        """
        self.project = project
        self.database = database
        self.collection = collection
        self.service_account_info = service_account_info
        self.service_account_file = service_account_file
        self.access_token = access_token
        self.id_map_field = id_map_field
        self.pk_map_field = pk_map_field
        self.etag_embed_field = etag_embed_field
        self.suppress_fields = suppress_fields
        self.nparams = nparams

        self._credential = None
        self._db_client = None
        self._adb_client = None
        self._collection_cache = dict()
        self._acollection_cache = dict()

    def init(self, context: Context | None = None) -> None:
        if self._db_client is not None:
            return

        from google.cloud.firestore import Client

        (
            self._credentials,
            self._db_client,
        ) = self._init_credentials_client(Client)

    async def ainit(self, context: Context | None = None) -> None:
        if self._adb_client is not None:
            return

        from google.cloud.firestore import AsyncClient

        (
            self._credentials,
            self._adb_client,
        ) = self._init_credentials_client(AsyncClient)

    def _init_credentials_client(self, client: Any):

        from verse.internal.storage_common._google import get_credentials

        credentials = get_credentials(
            self.service_account_info,
            self.service_account_file,
            self.access_token,
        )
        args = {
            "project": self.project,
            "database": self.database,
        }
        if credentials is not None:
            args["credentials"] = credentials
        args = args | self.nparams
        firestore_client: Any = client(**args)
        return credentials, firestore_client

    def _get_collection_name(
        self, op_parser: StoreOperationParser
    ) -> str | None:
        collection_name = (
            op_parser.get_operation_parsers()[0].get_collection_name()
            if op_parser.op_equals(StoreOperation.BATCH)
            else op_parser.get_collection_name()
        )
        db_collection = (
            collection_name if collection_name is not None else self.collection
        )
        return db_collection

    def _get_collections(
        self, op_parser: StoreOperationParser
    ) -> list[FirestoreCollection]:
        if op_parser.is_resource_op():
            return []
        if op_parser.op_equals(StoreOperation.TRANSACT):
            collections: list[FirestoreCollection] = []
            for single_op_parser in op_parser.get_operation_parsers():
                collections.extend(self._get_collections(single_op_parser))
            return collections
        db_collection = self._get_collection_name(op_parser)
        if db_collection is None:
            raise BadRequestError("Collection name must be specified")
        if db_collection in self._collection_cache:
            return [self._collection_cache[db_collection]]
        client = self._db_client.collection(db_collection)
        id_map_field = ParameterParser.get_collection_parameter(
            self.id_map_field, db_collection
        )
        pk_map_field = ParameterParser.get_collection_parameter(
            self.pk_map_field, db_collection
        )
        etag_embed_field = ParameterParser.get_collection_parameter(
            self.etag_embed_field, db_collection
        )
        col = FirestoreCollection(
            client,
            ClientHelper,
            self._db_client,
            id_map_field,
            pk_map_field,
            etag_embed_field,
            self.suppress_fields,
        )
        self._collection_cache[db_collection] = col
        return [col]

    async def _aget_collections(
        self, op_parser: StoreOperationParser
    ) -> list[FirestoreCollection]:
        if op_parser.is_resource_op():
            return []
        if op_parser.op_equals(StoreOperation.TRANSACT):
            collections: list[FirestoreCollection] = []
            for single_op_parser in op_parser.get_operation_parsers():
                collections.extend(
                    await self._aget_collections(single_op_parser)
                )
            return collections
        db_collection = self._get_collection_name(op_parser)
        if db_collection is None:
            raise BadRequestError("Collection name must be specified")
        if db_collection in self._acollection_cache:
            return [self._acollection_cache[db_collection]]
        client = self._adb_client.collection(db_collection)
        id_map_field = ParameterParser.get_collection_parameter(
            self.id_map_field, db_collection
        )
        pk_map_field = ParameterParser.get_collection_parameter(
            self.pk_map_field, db_collection
        )
        etag_embed_field = ParameterParser.get_collection_parameter(
            self.etag_embed_field, db_collection
        )
        col = FirestoreCollection(
            client,
            AsyncClientHelper,
            self._adb_client,
            id_map_field,
            pk_map_field,
            etag_embed_field,
            self.suppress_fields,
        )
        self._acollection_cache[db_collection] = col
        return [col]

    def _validate(self, op_parser: StoreOperationParser):
        if op_parser.op_equals(StoreOperation.BATCH):
            Validator.validate_batch(
                op_parser.get_operation_parsers(),
                allowed_ops=[StoreOperation.PUT, StoreOperation.DELETE],
                single_collection=True,
            )

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        self._validate(op_parser)
        collections = self._get_collections(op_parser)
        ncall, state = self._get_ncall(
            op_parser, collections, ResourceHelper(self._db_client)
        )
        if ncall is None:
            return super().run(operation)
        nresult = ncall.invoke()
        result = self._convert_nresult(
            nresult,
            state,
            op_parser,
            collections,
        )
        return Response(result=result, native=dict(result=nresult, call=ncall))

    async def arun(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        await self.ainit(context=context)
        op_parser = self.get_op_parser(operation)
        self._validate(op_parser)
        collections = await self._aget_collections(op_parser)
        ncall, state = self._get_ncall(
            op_parser, collections, AsyncResourceHelper(self._adb_client)
        )
        if ncall is None:
            return await super().arun(operation)
        nresult = await ncall.ainvoke()
        result = self._convert_nresult(
            nresult,
            state,
            op_parser,
            collections,
        )
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: StoreOperationParser,
        collections: list[FirestoreCollection],
        resource_helper: Any,
    ) -> tuple[NCall | None, dict | None]:
        if len(collections) == 1:
            converter = collections[0].converter
            client = collections[0].client
            helper = collections[0].helper
            db_client = collections[0].db_client
        call = None
        state = None
        nargs = op_parser.get_nargs()
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            args: dict = {
                "collection": self._get_collection_name(op_parser),
                "config": get_collection_config(op_parser),
                "nargs": nargs,
            }
            call = NCall(resource_helper.create_collection, args)
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            args = {
                "collection": self._get_collection_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(resource_helper.drop_collection, args)
        # LIST COLLECTION
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            args = {"nargs": nargs}
            call = NCall(resource_helper.list_collections, args)
        # GET
        if op_parser.op_equals(StoreOperation.GET):
            args = converter.convert_get(op_parser.get_key())
            call = NCall(
                client.document(args["id"]).get,
                None,
                nargs,
                {None: NotFoundError},
            )
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            args, state, func, id = converter.convert_put(
                op_parser.get_key(),
                op_parser.get_value(),
                op_parser.get_where(),
                op_parser.get_where_exists(),
            )
            if func == "set":
                call = NCall(
                    client.document(id).set,
                    args,
                    nargs,
                )
            elif func == "create":
                call = NCall(
                    client.document(id).create,
                    args,
                    nargs,
                    {AlreadyExists: PreconditionFailedError},
                )
            elif func == "helper":
                args["transaction"] = db_client.transaction()
                call = NCall(helper.put, args)
        # UPDATE
        elif op_parser.op_equals(StoreOperation.UPDATE):
            (args, state, func, id) = converter.convert_update(
                op_parser.get_key(),
                op_parser.get_set(),
                op_parser.get_where(),
                op_parser.get_where_exists(),
                op_parser.get_returning(),
            )
            if func == "update":
                call = NCall(
                    client.document(id).update,
                    args,
                    nargs,
                    {NotFound: NotFoundError},
                )
            elif func == "helper":
                args["transaction"] = db_client.transaction()
                call = NCall(helper.update, args)
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            args, func, id = converter.convert_delete(
                op_parser.get_key(),
                op_parser.get_where(),
                op_parser.get_where_exists(),
            )
            if func == "delete":
                call = NCall(
                    client.document(id).delete,
                    args,
                    nargs,
                    {NotFound: NotFoundError},
                )
            elif func == "helper":
                args["transaction"] = db_client.transaction()
                call = NCall(helper.delete, args)
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            args = converter.convert_query(
                client=client,
                select=op_parser.get_select(),
                where=op_parser.get_where(),
                order_by=op_parser.get_order_by(),
                limit=op_parser.get_limit(),
                offset=op_parser.get_offset(),
            )
            args = {"call": args["query"].stream, "args": None, "nargs": nargs}
            call = NCall(helper.query, args)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            args = converter.convert_count(
                client=client,
                where=op_parser.get_where(),
            )
            call = NCall(args["query"].get, None, nargs)
        # BATCH
        elif op_parser.op_equals(StoreOperation.BATCH):
            args, state = converter.convert_batch(
                op_parser.get_operation_parsers()
            )
            args["bulk_writer"] = db_client.bulk_writer()
            call = NCall(helper.batch, args, None)
        # TRANSACT
        elif op_parser.op_equals(StoreOperation.TRANSACT):
            args, state = OperationConverter.convert_transact(
                op_parser.get_operation_parsers(),
                [col.converter for col in collections],
            )
            args["transaction"] = resource_helper.db_client.transaction()
            call = NCall(resource_helper.transact, args, None)
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            args = {"nargs": nargs}
            call = NCall(resource_helper.close, args)
        return call, state

    def _convert_nresult(
        self,
        nresult: Any,
        state: dict | None,
        op_parser: StoreOperationParser,
        collections: list[FirestoreCollection],
    ) -> Any:
        if len(collections) == 1:
            processor = collections[0].processor
        result: Any = None
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            result = None
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            result = None
        # LIST COLLECTION
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            result = nresult
        # GET
        if op_parser.op_equals(StoreOperation.GET):
            if nresult is None or nresult.to_dict() is None:
                raise NotFoundError
            result = build_item_from_value(
                processor=processor,
                value=nresult.to_dict(),
                include_value=True,
            )
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            result = build_item_from_value(
                processor=processor,
                value=state["value"] if state is not None else None,
            )
        # UPDATE
        elif op_parser.op_equals(StoreOperation.UPDATE):
            if isinstance(nresult, dict):
                result = build_item_from_value(
                    processor=processor, value=nresult, include_value=True
                )
            else:
                key = op_parser.get_key()
                result = build_item_from_parts(
                    processor=processor,
                    key=key,
                    etag=(
                        state["etag"]
                        if state is not None and "etag" in state
                        else None
                    ),
                )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            result = None
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            items: list = []
            for item in nresult:
                items.append(
                    build_item_from_value(
                        processor=processor,
                        value=item.to_dict(),
                        include_value=True,
                    )
                )
            result = build_query_result(items)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            result = nresult[0][0].value
        # BATCH
        elif op_parser.op_equals(StoreOperation.BATCH):
            result = []
            if state is not None:
                for st in state["values"]:
                    if st is not None:
                        result.append(
                            build_item_from_value(
                                processor=processor,
                                value=st["value"],
                            )
                        )
                    else:
                        result.append(None)
        # TRANSACT
        elif op_parser.op_equals(StoreOperation.TRANSACT):
            result = []
            op_parsers = op_parser.get_operation_parsers()
            for i in range(0, len(op_parsers)):
                if state is not None:
                    st = state["values"][i]
                result.append(
                    self._convert_nresult(
                        nresult[i], st, op_parsers[i], [collections[i]]
                    )
                )
        return result


class OperationConverter:
    processor: ItemProcessor
    client: Any

    def __init__(self, processor: ItemProcessor, client: Any):
        self.processor = processor
        self.client = client

    @staticmethod
    def convert_transact(
        op_parsers: list[StoreOperationParser],
        converters: list[OperationConverter],
    ) -> tuple[dict, dict]:
        ops: list = []
        states: list = []
        for i in range(0, len(op_parsers)):
            op_parser = op_parsers[i]
            converter = converters[i]
            if op_parser.op_equals(StoreOperation.PUT):
                args, state, func, id = converter.convert_put(
                    op_parser.get_key(),
                    op_parser.get_value(),
                    op_parser.get_where(),
                    op_parser.get_where_exists(),
                )
                ops.append(
                    {
                        "op": StoreOperation.PUT,
                        "id": id,
                        "func": func,
                        "args": args,
                        "client": converter.client,
                        "processor": converter.processor,
                    }
                )
                states.append(state)
            elif op_parser.op_equals(StoreOperation.UPDATE):
                (args, state, func, id) = converter.convert_update(
                    op_parser.get_key(),
                    op_parser.get_set(),
                    op_parser.get_where(),
                    op_parser.get_where_exists(),
                    op_parser.get_returning(),
                )
                ops.append(
                    {
                        "op": StoreOperation.UPDATE,
                        "id": id,
                        "func": func,
                        "args": args,
                        "client": converter.client,
                        "processor": converter.processor,
                    }
                )
                states.append(state)
            elif op_parser.op_equals(StoreOperation.DELETE):
                args, func, id = converter.convert_delete(
                    op_parser.get_key(),
                    op_parser.get_where(),
                    op_parser.get_where_exists(),
                )
                ops.append(
                    {
                        "op": StoreOperation.DELETE,
                        "id": id,
                        "func": func,
                        "args": args,
                        "client": converter.client,
                        "processor": converter.processor,
                    }
                )
                states.append(None)
        return {"ops": ops}, {"values": states}

    def convert_batch(
        self, op_parsers: list[StoreOperationParser]
    ) -> tuple[dict, dict]:
        ops: list = []
        states: list = []
        for op_parser in op_parsers:
            if op_parser.op_equals(StoreOperation.PUT):
                args, state, func, id = self.convert_put(
                    op_parser.get_key(),
                    op_parser.get_value(),
                    op_parser.get_where(),
                    op_parser.get_where_exists(),
                )
                ops.append(
                    {
                        "op": StoreOperation.PUT,
                        "id": id,
                        "document": args["document_data"],
                    }
                )
                states.append(state)
            elif op_parser.op_equals(StoreOperation.DELETE):
                args, func, id = self.convert_delete(
                    op_parser.get_key(),
                    op_parser.get_where(),
                    op_parser.get_where_exists(),
                )
                ops.append({"op": StoreOperation.DELETE, "id": id})
                states.append(None)
        return {"ops": ops}, {"values": states}

    def convert_get(self, key: Value) -> dict:
        id = self.processor.get_id_from_key(key)
        return {"id": id}

    def convert_put(
        self,
        key: Value,
        value: dict,
        where: Expression | None,
        exists: bool | None,
    ) -> tuple[dict, dict | None, str, Value]:
        document = self.processor.add_embed_fields(value, key)
        if key is not None:
            id = self.processor.get_id_from_key(key)
        else:
            id = self.processor.get_id_from_value(document)
        if where is None:
            args: dict = {"document_data": document}
            func = "set"
        elif exists is False:
            args = {"document_data": document}
            func = "create"
        elif where is not None:
            args = {
                "id": id,
                "document_data": document,
                "where": where,
            }
            func = "helper"
        return args, {"value": document}, func, id

    def convert_update(
        self,
        key: Value,
        set: Update,
        where: Expression | None,
        exists: bool | None,
        returning: bool | None,
    ) -> tuple[dict, dict | None, str, Value]:
        state = None
        id = self.processor.get_id_from_key(key=key)
        if self.processor.needs_local_etag():
            etag = self.processor.generate_etag()
            set = self.processor.add_etag_update(set, etag)
            state = {"etag": etag}
        field_updates = self.convert_update_ops(set)
        if (where is None or exists is True) and returning is None:
            args: dict = {
                "field_updates": field_updates,
            }
            func = "update"
        elif where is not None or returning is not None:
            args = {
                "id": id,
                "field_updates": field_updates,
                "where": where,
                "update": set,
                "returning": returning,
            }
            func = "helper"
        return args, state, func, id

    def convert_delete(
        self, key: Value, where: Expression | None, exists: bool | None
    ) -> tuple[dict, str, Value]:
        id = self.processor.get_id_from_key(key=key)
        if where is None or exists is True:
            args = {"option": ExistsOption(exists=True)}
            func = "delete"
        else:
            args = {
                "id": id,
                "where": where,
            }
            func = "helper"
        return args, func, id

    def convert_query(
        self,
        client: Any,
        select: Select | None = None,
        where: Expression | None = None,
        order_by: OrderBy | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> dict:
        query = client
        if select is not None:
            query = self.convert_select(query, select)
        if where is not None:
            filter = self.convert_where(where)
            query = query.where(filter=filter)
        if order_by is not None:
            query = self.convert_order_by(query, order_by)
        if limit is not None:
            query = query.limit(limit)
        if offset is not None:
            query = query.offset(offset)
        return {"query": query}

    def convert_count(
        self, client: Any, where: Expression | None = None
    ) -> dict:
        query = client
        if where is not None:
            filter = self.convert_where(where)
            query = query.where(filter=filter)
        query = query.count()
        return {"query": query}

    def convert_field(self, field: Field | str) -> str:
        path: str
        if isinstance(field, str):
            path = field
        elif isinstance(field, Field):
            path = field.path
        return self.processor.resolve_field(path)

    def convert_where(self, expr: Expression | None) -> BaseFilter:
        if isinstance(expr, Function):
            return self.convert_func(expr)
        if isinstance(expr, Comparison):
            field = None
            field_value = None
            if isinstance(expr.lexpr, Field):
                field = self.convert_field(expr.lexpr)
                field_value = expr.rexpr
            elif isinstance(expr.rexpr, Field):
                field = self.convert_field(expr.rexpr)
                field_value = expr.lexpr
            else:
                raise BadRequestError(f"Comparison {expr} not supported")
            op = expr.op.value
            if expr.op == ComparisonOp.EQ:
                op = "=="
            if expr.op == ComparisonOp.IN:
                op = "in"
            elif expr.op == ComparisonOp.NIN:
                op = "not-in"
            if expr.op == ComparisonOp.BETWEEN and isinstance(
                field_value, list
            ):
                return AndFilter(
                    [
                        FieldFilter(field, ">=", field_value[0]),
                        FieldFilter(field, "<=", field_value[1]),
                    ],
                )
            return FieldFilter(field, op, field_value)
        if isinstance(expr, And):
            return AndFilter(
                [
                    self.convert_where(expr.lexpr),
                    self.convert_where(expr.rexpr),
                ]
            )
        if isinstance(expr, Or):
            return OrFilter(
                [
                    self.convert_where(expr.lexpr),
                    self.convert_where(expr.rexpr),
                ]
            )
        raise BadRequestError(f"Expression {expr!r} not supported")

    def convert_func(self, expr: Function) -> BaseFilter:
        namespace = expr.namespace
        name = expr.name
        args = expr.args
        if namespace == FunctionNamespace.BUILTIN:
            if name == StoreFunctionName.STARTS_WITH:
                field = self.convert_field(args[0])
                value = args[1]
                return AndFilter(
                    [
                        FieldFilter(field, ">=", value),
                        FieldFilter(field, "<=", str(value) + "~"),
                    ],
                )
            if name == StoreFunctionName.ARRAY_CONTAINS:
                field = self.convert_field(args[0])
                value = args[1]
                return FieldFilter(field, "array_contains", value)
            if name == StoreFunctionName.ARRAY_CONTAINS_ANY:
                field = self.convert_field(args[0])
                value = args[1]
                return FieldFilter(field, "array_contains_any", value)
        raise BadRequestError(f"Function {name} not supported")

    def convert_order_by(self, query, order_by: OrderBy):
        for term in order_by.terms:
            args = {"field_path": self.convert_field(term.field)}
            if term.direction is not None:
                if term.direction == OrderByDirection.ASC:
                    args["direction"] = "ASCENDING"
                elif term.direction == OrderByDirection.DESC:
                    args["direction"] = "DESCENDING"
            return query.order_by(**args)

    def convert_select(self, query, select: Select):
        if len(select.terms) == 0:
            return query
        field_paths = []
        for term in select.terms:
            field_path = self.convert_field(term.field)
            field_paths.append(field_path)
        return query.select(field_paths)

    def convert_update_ops(self, update: Update) -> dict:
        field_updates = {}
        for operation in update.operations:
            field = self.processor.resolve_field(operation.field)
            if operation.op == UpdateOp.PUT:
                field_updates[field] = operation.args[0]
            elif operation.op == UpdateOp.INSERT:
                field_updates[field] = operation.args[0]
            elif operation.op == UpdateOp.DELETE:
                field_updates[field] = firestore.DELETE_FIELD
            elif operation.op == UpdateOp.INCREMENT:
                field_updates[field] = firestore.Increment(operation.args[0])
            elif operation.op == UpdateOp.ARRAY_UNION:
                field_updates[field] = firestore.ArrayUnion(operation.args[0])
            elif operation.op == UpdateOp.ARRAY_REMOVE:
                field_updates[field] = firestore.ArrayRemove(operation.args[0])
            else:
                raise BadRequestError("Update operation not supported")
        return field_updates


class ResourceHelper:
    db_client: Any

    def __init__(self, db_client):
        self.db_client = db_client

    def create_collection(
        self,
        collection: str,
        config: DocumentCollectionConfig | None,
        nargs: Any,
    ):
        pass

    def drop_collection(self, collection: str, nargs: Any):
        batch_size = 100
        docs = self.db_client.collection(collection).list_documents(
            page_size=batch_size
        )
        deleted = 0

        for doc in docs:
            doc.delete()
            deleted = deleted + 1

        if deleted >= batch_size:
            return self.drop_collection(collection, nargs)

    def list_collections(self, nargs) -> Any:
        nresult = []
        response = NCall(self.db_client.collections, None, nargs).invoke()
        for item in response:
            nresult.append(item.id)
        return nresult

    def transact(self, transaction: Any, ops: list[dict]) -> Any:
        return ResourceHelper.run_transaction(transaction, ops)

    @transactional
    def run_transaction(transaction: Any, ops: list[dict]) -> Any:
        result: list = []
        documents = []
        for op in ops:
            client = op["client"]
            processor = op["processor"]
            if op["func"] == "helper":
                where = op["args"]["where"]
                snapshot = client.document(op["id"]).get(
                    transaction=transaction
                )
                document = snapshot.to_dict()
                if not QueryProcessor.eval_expr(
                    document, where, processor.resolve_field
                ):
                    raise ConflictError
                documents.append(document)
            else:
                documents.append(None)
        try:
            for i in range(0, len(ops)):
                op = ops[i]
                client = op["client"]
                processor = op["processor"]
                ref = client.document(op["id"])
                if op["op"] == StoreOperation.PUT:
                    if op["func"] == "create":
                        transaction.create(
                            reference=ref,
                            document_data=op["args"]["document_data"],
                        )
                    else:
                        transaction.set(
                            reference=ref,
                            document_data=op["args"]["document_data"],
                        )
                    result.append(None)
                elif op["op"] == StoreOperation.UPDATE:
                    transaction.update(
                        reference=ref,
                        field_updates=op["args"]["field_updates"],
                    )
                    if "returning" in op["args"]:
                        returning = op["args"]["returning"]
                        if returning is False:
                            result.append(documents[i])
                        elif returning is True:
                            result.append(
                                QueryProcessor.update_item(
                                    documents[i],
                                    op["args"]["update"],
                                    processor.resolve_field,
                                )
                            )
                        else:
                            result.append(None)
                elif op["op"] == StoreOperation.DELETE:
                    if op["func"] == "delete":
                        transaction.delete(
                            reference=ref, option=ExistsOption(exists=True)
                        )
                    else:
                        transaction.delete(reference=ref)
                    result.append(None)
        except Exception:
            raise ConflictError
        return result

    def close(self, nargs: Any) -> Any:
        pass


class AsyncResourceHelper:
    db_client: Any

    def __init__(self, db_client):
        self.db_client = db_client

    async def create_collection(
        self,
        collection: str,
        config: DocumentCollectionConfig | None,
        nargs: Any,
    ):
        pass

    async def drop_collection(self, collection: str, nargs: Any):
        batch_size = 100
        docs = self.db_client.collection(collection).list_documents(
            page_size=batch_size
        )
        deleted = 0

        async for doc in docs:
            await doc.delete()
            deleted = deleted + 1

        if deleted >= batch_size:
            return await self.drop_collection(collection, nargs)

    async def list_collections(self, nargs) -> Any:
        nresult = []
        response = NCall(self.db_client.collections, None, nargs).invoke()
        async for item in response:
            nresult.append(item.id)
        return nresult

    async def transact(self, transaction: Any, ops: list[dict]) -> Any:
        return await AsyncResourceHelper.run_transaction(transaction, ops)

    @async_transactional
    async def run_transaction(transaction: Any, ops: list[dict]) -> Any:
        result: list = []
        documents = []
        for op in ops:
            client = op["client"]
            processor = op["processor"]
            if op["func"] == "helper":
                where = op["args"]["where"]
                snapshot = await client.document(op["id"]).get(
                    transaction=transaction
                )
                document = snapshot.to_dict()
                if not QueryProcessor.eval_expr(
                    document, where, processor.resolve_field
                ):
                    raise ConflictError
                documents.append(document)
            else:
                documents.append(None)
        try:
            for i in range(0, len(ops)):
                op = ops[i]
                client = op["client"]
                processor = op["processor"]
                ref = client.document(op["id"])
                if op["op"] == StoreOperation.PUT:
                    if op["func"] == "create":
                        transaction.create(
                            reference=ref,
                            document_data=op["args"]["document_data"],
                        )
                    else:
                        transaction.set(
                            reference=ref,
                            document_data=op["args"]["document_data"],
                        )
                    result.append(None)
                elif op["op"] == StoreOperation.UPDATE:
                    transaction.update(
                        reference=ref,
                        field_updates=op["args"]["field_updates"],
                    )
                    if "returning" in op["args"]:
                        returning = op["args"]["returning"]
                        if returning is False:
                            result.append(documents[i])
                        elif returning is True:
                            result.append(
                                QueryProcessor.update_item(
                                    documents[i],
                                    op["args"]["update"],
                                    processor.resolve_field,
                                )
                            )
                        else:
                            result.append(None)
                elif op["op"] == StoreOperation.DELETE:
                    if op["func"] == "delete":
                        transaction.delete(
                            reference=ref, option=ExistsOption(exists=True)
                        )
                    else:
                        transaction.delete(reference=ref)
                    result.append(None)
        except Exception:
            raise ConflictError
        return result

    async def close(self, nargs: Any) -> Any:
        pass


class ClientHelper:
    client: Any
    processor: ItemProcessor

    def __init__(self, client: Any, processor: ItemProcessor):
        self.client = client
        self.processor = processor

    def put(self, transaction, id, document_data, where) -> dict:
        doc_ref = self.client.document(id)
        result, doc = ClientHelper.run_conditional_op_in_transaction(
            transaction,
            transaction.set,
            {"reference": doc_ref, "document_data": document_data},
            where,
            doc_ref,
            self.processor.resolve_field,
            True,
        )
        if result is None:
            if where is None:
                raise NotFoundError
            else:
                raise PreconditionFailedError
        if result is False:
            raise PreconditionFailedError
        return document_data

    def update(
        self,
        transaction,
        id,
        field_updates,
        where,
        update: Update,
        returning: bool | None,
    ) -> dict | None:
        doc_ref = self.client.document(id)
        result, doc = ClientHelper.run_conditional_op_in_transaction(
            transaction,
            transaction.update,
            {"reference": doc_ref, "field_updates": field_updates},
            where,
            doc_ref,
            self.processor.resolve_field,
            returning is None,
        )
        if result is None:
            if where is None:
                raise NotFoundError
            else:
                raise PreconditionFailedError
        if result is False:
            raise PreconditionFailedError
        if returning is None:
            return None
        if returning is False:
            return doc
        return QueryProcessor.update_item(
            doc, update, self.processor.resolve_field
        )

    def delete(self, transaction, id, where) -> None:
        doc_ref = self.client.document(id)
        result, doc = ClientHelper.run_conditional_op_in_transaction(
            transaction,
            transaction.delete,
            {"reference": doc_ref},
            where,
            doc_ref,
            self.processor.resolve_field,
            True,
        )
        if result is None:
            if where is None:
                raise NotFoundError
            else:
                raise PreconditionFailedError
        if result is False:
            raise PreconditionFailedError
        return None

    def query(self, call, args, nargs) -> Any:
        nresult = NCall(call, args, nargs).invoke()
        return nresult

    def batch(self, bulk_writer: Any, ops: list[dict]) -> Any:
        for op in ops:
            if op["op"] == StoreOperation.PUT:
                bulk_writer.set(self.client.document(op["id"]), op["document"])
            elif op["op"] == StoreOperation.DELETE:
                bulk_writer.delete(self.client.document(op["id"]))
        bulk_writer.close()

    def close(self, nargs: Any) -> Any:
        pass

    @transactional
    def run_conditional_op_in_transaction(
        transaction,
        op,
        args,
        where,
        doc_ref,
        special_field_resolver,
        project_get,
    ) -> tuple:
        if project_get:
            fields = QueryProcessor.extract_filter_fields(
                where, special_field_resolver
            )
            snapshot = doc_ref.get(field_paths=fields, transaction=transaction)
        else:
            snapshot = doc_ref.get(transaction=transaction)
        document = snapshot.to_dict()
        try:
            condition_result = QueryProcessor.eval_expr(
                document, where, special_field_resolver
            )
        except Exception:
            if document is None:
                return None, None
            raise
        if condition_result:
            op(**args)
            return True, document
        if document is None:
            return None, None
        return False, document


class AsyncClientHelper:
    client: Any
    processor: ItemProcessor

    def __init__(self, client: Any, processor: ItemProcessor):
        self.client = client
        self.processor = processor

    async def put(self, transaction, id, document_data, where) -> dict:
        doc_ref = self.client.document(id)
        (
            result,
            doc,
        ) = await AsyncClientHelper.run_conditional_op_in_transaction(
            transaction,
            transaction.set,
            {"reference": doc_ref, "document_data": document_data},
            where,
            doc_ref,
            self.processor.resolve_field,
            True,
        )
        if result is None:
            if where is None:
                raise NotFoundError
            else:
                raise PreconditionFailedError
        if result is False:
            raise PreconditionFailedError
        return document_data

    async def update(
        self,
        transaction,
        id,
        field_updates,
        where,
        update: Update,
        returning: bool | None,
    ) -> dict | None:
        doc_ref = self.client.document(id)
        (
            result,
            doc,
        ) = await AsyncClientHelper.run_conditional_op_in_transaction(
            transaction,
            transaction.update,
            {"reference": doc_ref, "field_updates": field_updates},
            where,
            doc_ref,
            self.processor.resolve_field,
            returning is None,
        )
        if result is None:
            if where is None:
                raise NotFoundError
            else:
                raise PreconditionFailedError
        if result is False:
            raise PreconditionFailedError
        if returning is None:
            return None
        if returning is False:
            return doc
        return QueryProcessor.update_item(
            doc, update, self.processor.resolve_field
        )

    async def delete(self, transaction, id, where) -> None:
        doc_ref = self.client.document(id)
        (
            result,
            doc,
        ) = await AsyncClientHelper.run_conditional_op_in_transaction(
            transaction,
            transaction.delete,
            {"reference": doc_ref},
            where,
            doc_ref,
            self.processor.resolve_field,
            True,
        )
        if result is None:
            if where is None:
                raise NotFoundError
            else:
                raise PreconditionFailedError
        if result is False:
            raise PreconditionFailedError
        return None

    async def query(self, call, args, nargs) -> Any:
        response = NCall(call, args, nargs).invoke()
        nresult = []
        async for item in response:
            nresult.append(item)
        return nresult

    async def batch(self, bulk_writer: Any, ops: list[dict]) -> Any:
        for op in ops:
            if op["op"] == StoreOperation.PUT:
                bulk_writer.set(self.client.document(op["id"]), op["document"])
            elif op["op"] == StoreOperation.DELETE:
                bulk_writer.delete(self.client.document(op["id"]))
        bulk_writer.close()

    async def close(self, nargs: Any) -> Any:
        pass

    @async_transactional
    async def run_conditional_op_in_transaction(
        transaction,
        op,
        args,
        where,
        doc_ref,
        special_field_resolver,
        project_get,
    ) -> tuple:
        if project_get:
            fields = QueryProcessor.extract_filter_fields(
                where, special_field_resolver
            )
            snapshot = await doc_ref.get(
                field_paths=fields, transaction=transaction
            )
        else:
            snapshot = await doc_ref.get(transaction=transaction)
        document = snapshot.to_dict()
        try:
            condition_result = QueryProcessor.eval_expr(
                document, where, special_field_resolver
            )
        except Exception:
            if document is None:
                return None, None
            raise
        if condition_result:
            op(**args)
            return True, document
        if document is None:
            return None, None
        return False, document


class FirestoreCollection:
    client: Any
    helper: Any
    db_client: Any
    converter: OperationConverter
    processor: ItemProcessor

    def __init__(
        self,
        client: Any,
        helper_type: Any,
        db_client: Any,
        id_map_field: str | None,
        pk_map_field: str | None,
        etag_embed_field: str | None,
        suppress_fields: list[str] | None,
    ) -> None:
        self.client = client
        self.processor = ItemProcessor(
            etag_embed_field=etag_embed_field,
            id_map_field=id_map_field,
            pk_map_field=pk_map_field,
            local_etag=True,
            suppress_fields=suppress_fields,
        )
        self.converter = OperationConverter(self.processor, client)
        self.helper = helper_type(client, self.processor)
        self.db_client = db_client



from typing import Any

from verse.core import Response
from verse.internal.storage_core import StoreComponent, StoreOperation
from verse.ql import Expression

from ._models import ConfigItem, ConfigKey, ConfigList


class ConfigStore(StoreComponent):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(
        self,
        key: str | dict | ConfigKey,
        **kwargs: Any,
    ) -> Response[ConfigItem]:
        """Get config value.

        Args:
            key: Config key.

        Returns:
            Config item with value.

        Raises:
            NotFoundError: Key not found.
        """
        operation = StoreOperation.get(
            key=key,
            **kwargs,
        )
        return self.run(operation)

    def put(
        self,
        key: str | dict | ConfigKey,
        value: str,
        **kwargs: Any,
    ) -> Response[ConfigItem]:
        """Put config value.

        Args:
            key: Config key.
            value: Config value.

        Returns:
            Config item.
        """
        operation = StoreOperation.put(
            key=key,
            value=value,
            **kwargs,
        )
        return self.run(operation)

    def delete(
        self,
        key: str | dict | ConfigKey,
        **kwargs: Any,
    ) -> Response[None]:
        """Delete config.

        Args:
            key: Config key.

        Returns:
           None

        Raises:
            NotFoundError: Key not found.
        """
        operation = StoreOperation.delete(
            key=key,
            **kwargs,
        )
        return self.run(operation)

    def query(
        self,
        where: str | Expression | None = None,
        **kwargs,
    ) -> Response[ConfigList]:
        """Query config.

        Args:
            where: Condition expression, defaults to None.

        Returns:
            List of config items.
        """
        operation = StoreOperation.query(
            where=where,
            **kwargs,
        )
        return self.run(operation)

    def count(
        self,
        where: str | None = None,
        **kwargs,
    ) -> Response[int]:
        """Count config.

        Args:
            where: Condition expression, defaults to None.

        Returns:
            Count of config items.
        """
        operation = StoreOperation.count(
            where=where,
            **kwargs,
        )
        return self.run(operation)

    async def aget(
        self,
        key: str | dict | ConfigKey,
        **kwargs: Any,
    ) -> Response[ConfigItem]:
        """Get config value.

        Args:
            key: Config key.

        Returns:
            Config item with value.

        Raises:
            NotFoundError: Key not found.
        """
        operation = StoreOperation.get(
            key=key,
            **kwargs,
        )
        return await self.arun(operation)

    async def aput(
        self,
        key: str | dict | ConfigKey,
        value: str,
        **kwargs: Any,
    ) -> Response[ConfigItem]:
        """Put config value.

        Args:
            key: Config key.
            value: Config value.

        Returns:
            Config item.
        """
        operation = StoreOperation.put(
            key=key,
            value=value,
            **kwargs,
        )
        return await self.arun(operation)

    async def adelete(
        self,
        key: str | dict | ConfigKey,
        **kwargs: Any,
    ) -> Response[None]:
        """Delete config.

        Args:
            key: Config key.

        Returns:
           None

        Raises:
            NotFoundError: Key not found.
        """
        operation = StoreOperation.delete(
            key=key,
            **kwargs,
        )
        return await self.arun(operation)

    async def aquery(
        self,
        where: str | Expression | None = None,
        **kwargs,
    ) -> Response[ConfigList]:
        """Query config.

        Args:
            where: Condition expression, defaults to None.

        Returns:
            List of config items.
        """
        operation = StoreOperation.query(
            where=where,
            **kwargs,
        )
        return await self.arun(operation)

    async def acount(
        self,
        where: str | Expression | None = None,
        **kwargs,
    ) -> Response[int]:
        """Count config.

        Args:
            where: Condition expression, defaults to None.

        Returns:
            Count of config items.
        """
        operation = StoreOperation.count(
            where=where,
            **kwargs,
        )
        return await self.arun(operation)

    async def aclose(
        self,
        **kwargs: Any,
    ) -> Response[None]:
        """Close the async client.

        Returns:
            None.
        """
        operation = StoreOperation.close(
            **kwargs,
        )
        return await self.arun(operation)



from typing import Any

from verse.core import Response
from verse.internal.storage_core import StoreComponent, StoreOperation
from verse.ql import Expression

from ._models import KeyValueItem, KeyValueKey, KeyValueList


class KeyValueStore(StoreComponent):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(
        self,
        key: str | dict | KeyValueKey,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[KeyValueItem]:
        """Get value.

        Args:
            key: Key.

        Returns:
            KV item with value.

        Raises:
            NotFoundError: Key not found.
        """
        return self._run_internal(StoreOperation.GET, locals())

    def put(
        self,
        key: str | dict | KeyValueKey,
        value: str | bytes,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[KeyValueItem]:
        """Put value.

        Args:
            key: Key.
            value: Value.
            where: Condition expression.

        Returns:
            KV item.
        """
        return self._run_internal(StoreOperation.PUT, locals())

    def delete(
        self,
        key: str | dict | KeyValueKey,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Delete key.

        Args:
            key: Key.

        Returns:
           None

        Raises:
            NotFoundError: Key not found.
        """
        return self._run_internal(StoreOperation.DELETE, locals())

    def query(
        self,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[KeyValueList]:
        """Query items.

        Args:
            where: Condition expression.

        Returns:
            List of KV items.
        """
        return self._run_internal(StoreOperation.QUERY, locals())

    def count(
        self,
        where: str | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[int]:
        """Count items.

        Args:
            where: Condition expression.

        Returns:
            Count of KV items.
        """
        return self._run_internal(StoreOperation.COUNT, locals())

    async def aget(
        self,
        key: str | dict | KeyValueKey,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[KeyValueItem]:
        """Get value.

        Args:
            key: Key.

        Returns:
            KV item with value.

        Raises:
            NotFoundError: Key not found.
        """
        return await self._arun_internal(StoreOperation.GET, locals())

    async def aput(
        self,
        key: str | dict | KeyValueKey,
        value: str | bytes,
        where: str | Expression | None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[KeyValueItem]:
        """Put value.

        Args:
            key: Key.
            value: Value.
            where: Condition expression.

        Returns:
            KV item.
        """
        return await self._arun_internal(StoreOperation.PUT, locals())

    async def adelete(
        self,
        key: str | dict | KeyValueKey,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Delete key.

        Args:
            key: Key.

        Returns:
           None

        Raises:
            NotFoundError: Key not found.
        """
        return await self._arun_internal(StoreOperation.DELETE, locals())

    async def aquery(
        self,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[KeyValueList]:
        """Query items.

        Args:
            where: Condition expression.

        Returns:
            List of KV items.
        """
        return await self._arun_internal(StoreOperation.QUERY, locals())

    async def acount(
        self,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[int]:
        """Count items.

        Args:
            where: Condition expression.

        Returns:
            Count of KV items.
        """
        return await self._arun_internal(StoreOperation.COUNT, locals())

    async def aclose(
        self,
        **kwargs: Any,
    ) -> Response[None]:
        """Close the async client.

        Returns:
            None.
        """
        return await self._arun_internal(StoreOperation.CLOSE, locals())



"""
Config Store on YAML file.
"""

__all__ = ["YamlFile"]

from typing import Any

import yaml

from verse.core import Context, Operation, Response
from verse.core.exceptions import BadRequestError, NotFoundError
from verse.internal.storage_core import StoreOperation, StoreProvider

from .._constants import DEFAULT_LABEL
from .._helper import normalize_label
from .._models import ConfigItem, ConfigKey


class YamlFile(StoreProvider):
    path: str = "config.yaml"
    flat: bool = False

    _items: dict
    _init: bool

    def __init__(
        self,
        path: str = "config.yaml",
        flat: bool = False,
        **kwargs,
    ):
        """Initialize.

        Args:
            path:
                YAML file path, defaults to "config.yaml".
            flat:
                A value indicating whether
                the YAML file is flat without any label hierarchy,
                defaults to False.
        """
        self.path = path
        self.flat = flat

        self._init = False

    def init(self, context: Context | None = None) -> None:
        if self._init:
            return

        items = yaml.safe_load(self.path)

        if items is None:
            self._items = dict()
        else:
            self._items = items
        self._init = True

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        # GET value
        if op_parser.op_equals(StoreOperation.GET):
            id = op_parser.get_id_as_str()
            label = normalize_label(op_parser.get_label())
            if self.flat:
                if label != DEFAULT_LABEL:
                    raise BadRequestError(
                        "Custom label in not support in flat config file"
                    )
                if id not in self._items:
                    raise NotFoundError
                value = self._items[id]
            else:
                if label not in self._items:
                    raise NotFoundError
                if id not in self._items[label]:
                    raise NotFoundError
                value = self._items[label][id]
            result = ConfigItem(
                key=ConfigKey(id=id, label=label),
                value=value,
            )
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            pass
        else:
            return super().run(operation)
        return Response(result=result)



from datetime import datetime

from verse.core import DataModel


class SecretKey(DataModel):
    """Secret key."""

    id: str
    """Secret id."""

    version: str | None = None
    """Secret version."""


class SecretProperties(DataModel):
    """Secret properties."""

    created_time: datetime | None = None
    """Created time."""


class SecretVersion(DataModel):
    """Secret version."""

    version: str
    "Version id."

    created_time: datetime | None = None
    """Created time."""


class SecretItem(DataModel):
    """Secret item."""

    key: SecretKey
    """Secret key."""

    value: str | None = None
    """Secret value."""

    metadata: dict | None = None
    """Secret metadata."""

    properties: SecretProperties | None = None
    """Secret properties."""

    versions: list[SecretVersion] | None = None
    """Secret versions."""


class SecretList(DataModel):
    """Secret list."""

    items: list[SecretItem]
    """List of secret items."""



import json


class Comparator:
    @staticmethod
    def contains(item1: dict, item2: dict) -> bool:
        for key, value in item1.items():
            if key not in item2:
                return False
            if isinstance(value, dict):
                if not Comparator.contains(value, item2[key]):
                    return False
            elif isinstance(value, list):
                if not isinstance(item2[key], list):
                    return False
                if len(value) != len(item2[key]):
                    return False
                for i1, i2 in zip(value, item2[key]):
                    if isinstance(i1, dict) and isinstance(i2, dict):
                        if not Comparator.contains(i1, i2):
                            return False
                    elif i1 != i2:
                        return False
            else:
                if value != item2[key]:
                    return False
        return True

    @staticmethod
    def equals(item1: dict, item2: dict) -> bool:
        str1 = json.dumps(item1, sort_keys=True)
        str2 = json.dumps(item2, sort_keys=True)
        return str1 == str2



from verse.core.exceptions import (
    BadRequestError,
    ConflictError,
    NotFoundError,
    NotModified,
    PreconditionFailedError,
)
from verse.internal.storage_core import StoreFunction

from ._models import (
    ObjectCollectionConfig,
    ObjectItem,
    ObjectKey,
    ObjectList,
    ObjectProperties,
    ObjectQueryConfig,
    ObjectSource,
    ObjectStoreClass,
    ObjectTransferConfig,
    ObjectVersion,
)
from .component import ObjectOperation, ObjectStore

__all__ = [
    "ObjectCollectionConfig",
    "ObjectItem",
    "ObjectKey",
    "ObjectList",
    "ObjectOperation",
    "ObjectProperties",
    "ObjectQueryConfig",
    "ObjectSource",
    "ObjectStore",
    "ObjectTransferConfig",
    "ObjectVersion",
    "ObjectStoreClass",
    "StoreFunction",
    "BadRequestError",
    "ConflictError",
    "NotFoundError",
    "NotModified",
    "PreconditionFailedError",
]



from typing import Any

from verse.core import Component
from verse.core.exceptions import NotSupportedError

from ._operation import StoreOperation


class StoreComponent(Component):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __getitem__(self, key: str) -> Any:
        op = StoreOperation.get(key=key)
        try:
            response = self.run(op)
            if response is not None and response.result is not None:
                return response.result.value
            return None
        except NotSupportedError:
            return None

    def __setitem__(self, key: str, value: Any) -> None:
        op = StoreOperation.put(key=key, value=value)
        self.run(op)



"""
Config Store on Environment variables.
"""

__all__ = ["Environment"]

import os
from typing import Any

from verse.core import Context, Operation, Response
from verse.core.exceptions import BadRequestError, NotFoundError
from verse.internal.storage_core import StoreOperation, StoreProvider

from .._constants import DEFAULT_LABEL
from .._helper import normalize_label
from .._models import ConfigItem, ConfigKey


class Environment(StoreProvider):
    def __init__(self, **kwargs):
        """Initialize."""
        pass

    def init(self, context: Context | None = None) -> None:
        pass

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        # GET value
        if op_parser.op_equals(StoreOperation.GET):
            id = op_parser.get_id_as_str()
            label = normalize_label(op_parser.get_label())
            if label != DEFAULT_LABEL:
                raise BadRequestError(
                    "Custom label in not support in environment"
                )
            val = os.environ.get(id)
            if val is None:
                raise NotFoundError
            result = ConfigItem(
                key=ConfigKey(id=id, label=label),
                value=val,
            )
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            id = op_parser.get_id_as_str()
            label = normalize_label(op_parser.get_label())
            value = str(op_parser.get_value())
            if label != DEFAULT_LABEL:
                raise BadRequestError(
                    "Custom label in not support in environment"
                )
            os.environ[id] = value
            result = ConfigItem(key=ConfigKey(id=id, label=label))
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            id = op_parser.get_id_as_str()
            label = normalize_label(op_parser.get_label())
            if label != DEFAULT_LABEL:
                raise BadRequestError(
                    "Custom label in not support in environment"
                )
            try:
                os.environ.pop(id)
            except KeyError:
                raise NotFoundError
            result = None
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            result = None
        else:
            return super().run(operation)
        return Response(result=result)



"""
Vector Store on Qdrant.
"""

from __future__ import annotations

__all__ = ["Qdrant"]

import copy
from typing import Any

from qdrant_client import models

from verse.core import Context, NCall, Operation, Response
from verse.core.exceptions import BadRequestError, NotFoundError
from verse.internal.storage_core import (
    ItemProcessor,
    QueryProcessor,
    StoreFunctionName,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
    Validator,
)
from verse.ql import (
    And,
    Comparison,
    ComparisonOp,
    Expression,
    Field,
    Function,
    FunctionNamespace,
    Not,
    Or,
    OrderBy,
    Select,
)

from .._helper import get_collection_config, is_metadata_field, is_value_field
from .._models import (
    VectorCollectionConfig,
    VectorIndexMetric,
    VectorItem,
    VectorKey,
    VectorList,
    VectorProperties,
    VectorSearchArgs,
    VectorValue,
)

DEFAULT_LIMIT = 100


class Qdrant(StoreProvider):
    location: str | None
    url: str | None
    api_key: str | None
    host: str | None
    port: int | None
    path: str | None
    collection: str | None
    vector_field: str
    sparse_vector_field: str | None
    content_payload_field: str | None
    nparams: dict[str, Any]

    _client: Any
    _aclient: Any
    _processor: ItemProcessor

    _collection_cache: dict[str, QdrantCollection]
    _acollection_cache: dict[str, QdrantCollection]

    def __init__(
        self,
        location: str | None = None,
        url: str | None = None,
        api_key: str | None = None,
        host: str | None = None,
        port: int | None = 6333,
        path: str | None = None,
        collection: str | None = None,
        vector_field: str = "vector",
        sparse_vector_field: str | None = "sparse_vector",
        content_payload_field: str | None = "_content",
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            location:
                Qdrant location.
            url:
                Qdrant url.
            api_key:
                Qdrant api key.
            host:
                Qdrant host.
            port:
                Qdrant port.
            path:
                Qdrant path.
            collection:
                Qdrant collection mapped to vector store collection.
            vector_field:
                Vector field name. Defaults to "vector".
            sparse_vector_field:
                Sparse vector field name. Defaults to "sparse_vector".
            content_payload_field:
                Payload field where content is stored.
                Qdrant doesn't support a separate content field.
                Defaults to "_content".
            nparams:
                Native parameters to Qdrant client.
        """
        self.location = location
        self.url = url
        self.api_key = api_key
        self.host = host
        self.port = port
        self.path = path
        self.collection = collection
        self.vector_field = vector_field
        self.sparse_vector_field = sparse_vector_field
        self.content_payload_field = content_payload_field
        self.nparams = nparams

        self._client = None
        self._aclient = None
        self._processor = ItemProcessor()
        self._collection_cache = dict()
        self._acollection_cache = dict()

    def init(self, context: Context | None = None) -> None:
        if self._client is not None:
            return

        from qdrant_client import QdrantClient

        self._client = QdrantClient(
            location=self.location,
            url=self.url,
            api_key=self.api_key,
            host=self.host,
            port=self.port,
            path=self.path,
            **self.nparams,
        )

    async def ainit(self, context: Context | None = None) -> None:
        if self._aclient is not None:
            return

        from qdrant_client import AsyncQdrantClient

        self._aclient = AsyncQdrantClient(
            location=self.location,
            url=self.url,
            api_key=self.api_key,
            host=self.host,
            port=self.port,
            path=self.path,
            **self.nparams,
        )

    def _get_collection_name(
        self, op_parser: StoreOperationParser
    ) -> str | None:
        collection_name = (
            op_parser.get_operation_parsers()[0].get_collection_name()
            if op_parser.op_equals(StoreOperation.BATCH)
            or op_parser.op_equals(StoreOperation.TRANSACT)
            else op_parser.get_collection_name()
        )
        collection = (
            collection_name if collection_name is not None else self.collection
        )
        return collection

    def _get_collection(
        self, op_parser: StoreOperationParser
    ) -> QdrantCollection | None:
        if op_parser.is_resource_op():
            return None
        collection_name = self._get_collection_name(op_parser)
        if collection_name is None:
            raise BadRequestError("Collection name must be specified")
        if collection_name in self._collection_cache:
            return self._collection_cache[collection_name]
        description = self._client.get_collection(collection_name)
        dimension = description.config.params.vectors[self.vector_field].size
        col = QdrantCollection(
            self._client,
            collection_name,
            dimension,
            self.vector_field,
            self.sparse_vector_field,
            self.content_payload_field,
        )
        self._collection_cache[collection_name] = col
        return col

    async def _aget_collection(
        self, op_parser: StoreOperationParser
    ) -> QdrantCollection | None:
        if op_parser.is_resource_op():
            return None
        collection_name = self._get_collection_name(op_parser)
        if collection_name is None:
            raise BadRequestError("Collection name must be specified")
        if collection_name in self._acollection_cache:
            return self._acollection_cache[collection_name]
        description = await self._aclient.get_collection(collection_name)
        dimension = description.config.params.vectors[self.vector_field].size
        col = QdrantCollection(
            self._aclient,
            collection_name,
            dimension,
            self.vector_field,
            self.sparse_vector_field,
            self.content_payload_field,
        )
        self._acollection_cache[collection_name] = col
        return col

    def _validate(self, op_parser: StoreOperationParser):
        if op_parser.op_equals(StoreOperation.BATCH):
            Validator.validate_batch(
                op_parser.get_operation_parsers(),
                allowed_ops=[
                    StoreOperation.PUT,
                    StoreOperation.DELETE,
                    StoreOperation.GET,
                ],
                single_collection=True,
            )
        elif op_parser.op_equals(StoreOperation.TRANSACT):
            Validator.validate_transact(
                op_parser.get_operation_parsers(),
                allowed_ops=[
                    StoreOperation.PUT,
                    StoreOperation.DELETE,
                    StoreOperation.GET,
                ],
                single_collection=True,
            )

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        self._validate(op_parser)
        collection = self._get_collection(op_parser)
        ncall, state = self._get_ncall(
            op_parser,
            collection,
            ResourceHelper(self._client),
        )
        if ncall is None:
            return super().run(operation)
        nresult = ncall.invoke()
        result = self._convert_nresult(nresult, state, op_parser, collection)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    async def arun(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        await self.ainit(context=context)
        op_parser = self.get_op_parser(operation)
        self._validate(op_parser)
        collection = await self._aget_collection(op_parser)
        ncall, state = self._get_ncall(
            op_parser,
            collection,
            AsyncResourceHelper(self._aclient),
        )
        if ncall is None:
            return await super().arun(operation)
        nresult = await ncall.ainvoke()
        result = self._convert_nresult(nresult, state, op_parser, collection)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: StoreOperationParser,
        collection: QdrantCollection | None,
        resource_helper: Any,
    ) -> tuple[NCall | None, dict | None]:
        if collection is not None:
            converter = collection.converter
            client = collection.client
        call = None
        state = None
        nargs = op_parser.get_nargs()
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            args: dict[Any, Any] | None = {
                "collection": self._get_collection_name(op_parser),
                "config": get_collection_config(op_parser),
                "vector_field": self.vector_field,
                "sparse_vector_field": self.sparse_vector_field,
                "nargs": nargs,
            }
            call = NCall(resource_helper.create_collection, args)
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            args = {
                "collection": self._get_collection_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(resource_helper.drop_collection, args)
        # LIST COLLECTION
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            args = {"nargs": nargs}
            call = NCall(resource_helper.list_collections, args)
        # GET
        elif op_parser.op_equals(StoreOperation.GET):
            args = converter.convert_get(op_parser.get_id_as_str())
            call = NCall(client.retrieve, args, nargs)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            args = converter.convert_put(
                op_parser.get_id_as_str(),
                op_parser.get_vector_value(),
                op_parser.get_metadata(),
            )
            call = NCall(client.upsert, args, nargs)
        # UPDATE value
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_value_update()
        ):
            args = converter.convert_update_value(
                id=op_parser.get_id_as_str(),
                value=op_parser.get_updated_value(),
            )
            call = NCall(client.update_vectors, args, nargs)
        # UPDATE metadata
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_metadata_update()
        ):
            args = converter.convert_update_metadata(
                id=op_parser.get_id_as_str(),
                metadata=op_parser.get_updated_metadata(),
            )
            call = NCall(client.set_payload, args, nargs)
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            args = converter.convert_delete(
                op_parser.get_id_as_str_or_none(), op_parser.get_where()
            )
            call = NCall(
                client.delete,
                args,
                nargs,
            )
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            args, func = converter.convert_query(
                search=op_parser.get_search_as_function(),
                where=op_parser.get_where(),
                select=op_parser.get_select(),
                order_by=op_parser.get_order_by(),
                limit=op_parser.get_limit(),
                offset=op_parser.get_offset(),
            )
            if func == "search":
                call = NCall(client.search, args, nargs)
            elif func == "search_batch":
                call = NCall(client.search_batch, args, nargs)
            state = {"func": func}
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            args = converter.convert_count(
                where=op_parser.get_where(),
            )
            call = NCall(client.count, args, nargs)
        # BATCH or TRANSACT
        elif op_parser.op_equals(StoreOperation.BATCH) or op_parser.op_equals(
            StoreOperation.TRANSACT
        ):
            args, func = converter.convert_batch(
                op_parser.get_operation_parsers()
            )
            if func == "get":
                call = NCall(client.retrieve, args, nargs)
            elif func == "put":
                call = NCall(client.upsert, args, nargs)
            elif func == "delete":
                call = NCall(client.delete, args, nargs)
            state = {"func": func}
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            args = {"nargs": nargs}
            call = NCall(resource_helper.close, args)
        return call, state

    def _convert_nresult(
        self,
        nresult: Any,
        state: dict | None,
        op_parser: StoreOperationParser,
        collection: QdrantCollection | None,
    ) -> Any:
        if collection is not None:
            processor = collection.processor
        result: Any = None
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            result = None
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            result = None
        # LIST COLLECTION
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            result = [c.name for c in nresult.collections]
        # GET
        elif op_parser.op_equals(StoreOperation.GET):
            if len(nresult) == 0:
                raise NotFoundError
            result = self._convert_to_item(nresult[0], processor)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            result = VectorItem(
                key=VectorKey(id=op_parser.get_id_as_str()),
            )
        # UPDATE value
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_value_update()
        ):
            result = VectorItem(key=VectorKey(id=op_parser.get_id_as_str()))
        # UPDATE metadata
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_metadata_update()
        ):
            result = VectorItem(
                key=VectorKey(id=op_parser.get_id_as_str()),
                metadata=op_parser.get_updated_metadata(),
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            result = None
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            items = []
            func = state["func"] if state is not None else None
            if func == "search":
                for item in nresult:
                    items.append(self._convert_to_item(item, processor))
                items = QueryProcessor.order_items(
                    items,
                    op_parser.get_order_by(),
                    processor.resolve_root_field,
                )
            elif func == "search_batch":
                limit = op_parser.get_limit()
                items = self._merge_batch(
                    nresult,
                    limit if limit is not None else DEFAULT_LIMIT,
                    processor,
                )
                items = QueryProcessor.order_items(
                    items,
                    op_parser.get_order_by(),
                    processor.resolve_root_field,
                )
            result = VectorList(items=items)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            result = nresult.count
        # BATCH or TRANSACT
        elif op_parser.op_equals(StoreOperation.BATCH) or op_parser.op_equals(
            StoreOperation.TRANSACT
        ):
            result = []
            func = state["func"] if state is not None else None
            if func == "get":
                for item in nresult:
                    result.append(self._convert_to_item(item, processor))
            elif func == "put":
                for op_parser in op_parser.get_operation_parsers():
                    result.append(
                        VectorItem(key=VectorKey(id=op_parser.get_id_as_str()))
                    )
            elif func == "delete":
                for op_parser in op_parser.get_operation_parsers():
                    result.append(None)
        return result

    def _merge_batch(
        self, batch_items: list, limit: int, processor: ItemProcessor
    ) -> list:
        unique_items: dict = dict()
        for nitems in batch_items:
            for nitem in nitems:
                item = self._convert_to_item(nitem, processor)
                if item.key.id in unique_items:
                    s1 = (
                        item.properties.score
                        if item.properties is not None
                        and item.properties.score is not None
                        else 0
                    )
                    dict_item = unique_items[item.key.id]
                    s2 = (
                        dict_item.properties.score
                        if dict_item.properties is not None
                        and dict_item.properties.score is not None
                        else 0
                    )
                    if s1 > s2:
                        unique_items[item.key.id] = item
                else:
                    unique_items[item.key.id] = item
        return sorted(
            list(unique_items.values())[0:limit],
            key=lambda x: x.properties.score,
            reverse=True,
        )

    def _convert_to_item(
        self, nitem: Any, processor: ItemProcessor
    ) -> VectorItem:
        content = None
        metadata: dict | None = None
        key = VectorKey(id=nitem.id)
        metadata = nitem.payload
        if metadata is not None:
            if (
                self.content_payload_field is not None
                and self.content_payload_field in metadata
            ):
                content = metadata[self.content_payload_field]
                metadata.pop(self.content_payload_field)
                if len(metadata) == 0:
                    metadata = None
        vector = None
        sparse_vector = None
        if nitem.vector is not None:
            if self.vector_field in nitem.vector:
                vector = nitem.vector[self.vector_field]
            if (
                self.sparse_vector_field is not None
                and self.sparse_vector_field in nitem.vector
            ):
                sparse_vector = dict()
                nsparse_vector = nitem.vector[self.sparse_vector_field]
                for i in range(0, len(nsparse_vector.indices)):
                    sparse_vector[nsparse_vector.indices[i]] = (
                        nsparse_vector.values[i]
                    )
        value = VectorValue(
            vector=vector,
            sparse_vector=sparse_vector,
            content=content,
        )
        properties = None
        if isinstance(nitem, models.ScoredPoint):
            properties = VectorProperties(score=nitem.score)
        return VectorItem(
            key=key, value=value, metadata=metadata, properties=properties
        )


class ResourceHelper:
    client: Any

    def __init__(self, client: Any):
        self.client = client

    def create_collection(
        self,
        collection: str,
        config: VectorCollectionConfig | None,
        vector_field: str,
        sparse_vector_field: str | None,
        nargs: Any,
    ):
        args = OperationConverter.convert_create_collection(
            collection, config, vector_field, sparse_vector_field
        )
        NCall(self.client.create_collection, args, nargs).invoke()

    def drop_collection(self, collection: str, nargs: Any):
        args = {
            "collection_name": collection,
        }
        NCall(self.client.delete_collection, args, nargs).invoke()

    def list_collections(self, nargs) -> Any:
        return NCall(self.client.get_collections, None, nargs).invoke()

    def close(self, nargs: Any) -> Any:
        pass


class AsyncResourceHelper:
    client: Any

    def __init__(self, client: Any):
        self.client = client

    async def create_collection(
        self,
        collection: str,
        config: VectorCollectionConfig | None,
        vector_field: str,
        sparse_vector_field: str | None,
        nargs: Any,
    ):
        args = OperationConverter.convert_create_collection(
            collection, config, vector_field, sparse_vector_field
        )
        await NCall(self.client.create_collection, args, nargs).ainvoke()

    async def drop_collection(self, collection: str, nargs: Any):
        args = {
            "collection_name": collection,
        }
        await NCall(self.client.delete_collection, args, nargs).ainvoke()

    async def list_collections(self, nargs) -> Any:
        return await NCall(self.client.get_collections, None, nargs).ainvoke()

    async def close(self, nargs: Any) -> Any:
        pass


class OperationConverter:
    processor: ItemProcessor
    collection: str | None
    dimension: int
    vector_field: str
    sparse_vector_field: str | None
    content_payload_field: str | None

    def __init__(
        self,
        processor: ItemProcessor,
        collection: str | None,
        dimension: int,
        vector_field: str,
        sparse_vector_field: str | None,
        content_payload_field: str | None,
    ) -> None:
        self.processor = processor
        self.collection = collection
        self.dimension = dimension
        self.vector_field = vector_field
        self.sparse_vector_field = sparse_vector_field
        self.content_payload_field = content_payload_field

    @staticmethod
    def convert_create_collection(
        collection: str,
        config: VectorCollectionConfig | None,
        vector_field: str,
        sparse_vector_field: str | None,
    ) -> dict:
        args: dict = {"collection_name": collection}
        metric_mapping = {
            VectorIndexMetric.DOTPRODUCT: models.Distance.DOT,
            VectorIndexMetric.COSINE: models.Distance.COSINE,
            VectorIndexMetric.EUCLIDEAN: models.Distance.EUCLID,
            VectorIndexMetric.MANHATTAN: models.Distance.MANHATTAN,
        }
        vector_index = config.vector_index if config is not None else None
        if vector_index is not None:
            dimension = vector_index.dimension
            metric = metric_mapping[vector_index.metric]
            nconfig = (
                vector_index.nconfig
                if vector_index.nconfig is not None
                else {}
            )
        else:
            dimension = 4
            metric = models.Distance.DOT
            nconfig = {}

        args["vectors_config"] = {
            vector_field: models.VectorParams(
                size=dimension, distance=metric, **nconfig
            )
        }

        if sparse_vector_field is not None:
            args["sparse_vectors_config"] = {
                sparse_vector_field: models.SparseVectorParams()
            }

        return args

    def convert_batch(
        self, op_parsers: list[StoreOperationParser]
    ) -> tuple[dict, str]:
        batch_get = False
        batch_put = False
        batch_delete = False
        for op_parser in op_parsers:
            if op_parser.op_equals(StoreOperation.GET):
                batch_get = True
            elif op_parser.op_equals(StoreOperation.PUT):
                batch_put = True
            elif op_parser.op_equals(StoreOperation.DELETE):
                batch_delete = True
        if (
            (batch_get and batch_put)
            or (batch_put and batch_delete)
            or (batch_delete and batch_get)
        ):
            raise BadRequestError(
                "Batch operation supports only one operation type"
            )
        if batch_get:
            return self.convert_batch_get(op_parsers), "get"
        elif batch_put:
            return self.convert_batch_put(op_parsers), "put"
        elif batch_delete:
            return self.convert_batch_delete(op_parsers), "delete"
        raise BadRequestError("Batch operation not supported")

    def convert_batch_get(
        self, op_parsers: list[StoreOperationParser]
    ) -> dict:
        ids = []
        for op_parser in op_parsers:
            ids.append(op_parser.get_id_as_str())
        return {
            "collection_name": self.collection,
            "ids": ids,
            "with_vectors": True,
        }

    def convert_batch_put(
        self, op_parsers: list[StoreOperationParser]
    ) -> dict:
        vectors = []
        for op_parser in op_parsers:
            vectors.append(
                self.convert_put_vector(
                    op_parser.get_id_as_str(),
                    op_parser.get_vector_value(),
                    op_parser.get_metadata(),
                )
            )
        return {
            "collection_name": self.collection,
            "points": vectors,
        }

    def convert_batch_delete(
        self, op_parsers: list[StoreOperationParser]
    ) -> dict:
        ids: list = []
        for op_parser in op_parsers:
            ids.append(op_parser.get_id_as_str())
        return {
            "collection_name": self.collection,
            "points_selector": models.PointIdsList(
                points=ids,
            ),
        }

    def convert_get(self, id: str) -> dict:
        return {
            "collection_name": self.collection,
            "ids": [id],
            "with_vectors": True,
        }

    def convert_put(self, id: str, value: dict, metadata: dict | None) -> dict:
        return {
            "collection_name": self.collection,
            "points": [self.convert_put_vector(id, value, metadata)],
        }

    def convert_put_vector(
        self, id: str, value: dict, metadata: dict | None
    ) -> Any:
        vector_value = VectorValue(**value)
        args: dict = {}
        args[self.vector_field] = vector_value.vector
        if (
            self.sparse_vector_field is not None
            and vector_value.sparse_vector is not None
        ):
            args[self.sparse_vector_field] = self.convert_sparse_vector(
                vector_value.sparse_vector
            )
        payload: dict | None = None
        if metadata is None:
            if (
                vector_value.content is not None
                and self.content_payload_field is not None
            ):
                payload = {self.content_payload_field: vector_value.content}
        else:
            payload = copy.deepcopy(metadata)
            if (
                vector_value.content is not None
                and self.content_payload_field is not None
            ):
                payload[self.content_payload_field] = vector_value.content
        return models.PointStruct(id=id, vector=args, payload=payload)

    def convert_update_value(self, id: str, value: dict) -> dict:
        if self.sparse_vector_field is not None:
            raise BadRequestError("Sparse vector update is not supported")
        if self.content_payload_field is not None:
            raise BadRequestError("Content update is not supported")
        args: dict = {"collection_name": self.collection}
        vector_value = VectorValue(**value)
        vector: dict = {self.vector_field: vector_value.vector}
        args["points"] = [models.PointVectors(id=id, vector=vector)]
        return args

    def convert_update_metadata(self, id: str, metadata: dict | None) -> dict:
        return {
            "collection_name": self.collection,
            "payload": metadata,
            "points": [id],
        }

    def convert_delete(self, id: str | None, where: Expression) -> dict:
        if id is None and where is None:
            raise BadRequestError("Id or where should be provided for delete")
        if id is not None and where is None:
            return {
                "collection_name": self.collection,
                "points_selector": models.PointIdsList(
                    points=[id],
                ),
            }
        if where is not None and id is None:
            return {
                "collection_name": self.collection,
                "points_selector": models.FilterSelector(
                    filter=models.Filter(must=[self.convert_expr(where)])
                ),
            }
        raise BadRequestError(
            "Both id and where cannot be provided for delete"
        )

    def convert_query(
        self,
        search: Function | None,
        where: Expression,
        select: Select | None,
        order_by: OrderBy | None,
        limit: int | None,
        offset: int | None,
    ) -> tuple[dict, str]:
        search_args = self.convert_search(search)
        filter = None
        if where is not None:
            filter = models.Filter(must=[self.convert_expr(where)])
        limit_arg = {"limit": DEFAULT_LIMIT if limit is None else limit}
        if isinstance(search_args, list):
            return {
                "collection_name": self.collection,
                "requests": [
                    models.SearchRequest(
                        **(
                            self.convert_select(select, True)
                            | search_args[0]
                            | limit_arg
                            | {
                                "filter": filter,
                                "offset": offset,
                            }
                        )
                    ),
                    models.SearchRequest(
                        **(
                            self.convert_select(select, True)
                            | search_args[1]
                            | limit_arg
                            | {
                                "filter": filter,
                                "offset": offset,
                            }
                        )
                    ),
                ],
            }, "search_batch"
        else:
            return (
                search_args
                | self.convert_select(select, False)
                | limit_arg
                | {
                    "collection_name": self.collection,
                    "query_filter": filter,
                    "offset": offset,
                }
            ), "search"

    def convert_select(self, select: Select | None, is_batch: bool) -> dict:
        include_value = False
        include_metadata = False
        if select is None or len(select.terms) == 0:
            include_value = True
            include_metadata = True
        else:
            for term in select.terms:
                if is_value_field(term.field):
                    # include metadata since content is stored there
                    # filter it out when returning
                    # optimize later
                    include_value = True
                    include_metadata = True
                elif is_metadata_field(term.field):
                    include_metadata = True
        if is_batch:
            return {
                "with_vector": include_value,
                "with_payload": include_metadata,
            }
        else:
            return {
                "with_vectors": include_value,
                "with_payload": include_metadata,
            }

    def convert_search(self, search: Function | None) -> Any:
        if search is None:
            return {
                "query_vector": models.NamedVector(
                    name=self.vector_field, vector=[0.0] * self.dimension
                )
            }
        namespace = search.namespace
        name = search.name
        if namespace == FunctionNamespace.BUILTIN:
            if name == StoreFunctionName.VECTOR_SEARCH:
                args = VectorSearchArgs(**search.named_args)
                if args.vector is not None and args.sparse_vector is None:
                    return self.convert_vector_search(args.vector, False)
                elif args.vector is None and args.sparse_vector is not None:
                    return self.convert_sparse_vector_search(
                        args.sparse_vector, False
                    )
                elif (
                    args.vector is not None and args.sparse_vector is not None
                ):
                    return [
                        self.convert_vector_search(args.vector, True),
                        self.convert_sparse_vector_search(
                            args.sparse_vector, True
                        ),
                    ]

        raise BadRequestError("Search function not supported")

    def convert_vector_search(
        self, vector: list[float], is_batch: bool
    ) -> Any:
        key = "vector" if is_batch else "query_vector"
        return {key: models.NamedVector(name=self.vector_field, vector=vector)}

    def convert_sparse_vector_search(
        self, sparse_vector: dict, is_batch: bool
    ) -> Any:
        key = "vector" if is_batch else "query_vector"
        if self.sparse_vector_field is not None:
            return {
                key: models.NamedSparseVector(
                    name=self.sparse_vector_field,
                    vector=self.convert_sparse_vector(sparse_vector),
                )
            }
        raise BadRequestError("Sparse vector search not supported")

    def convert_sparse_vector(
        self, sparse_vector: dict[int, float] | None
    ) -> Any:
        if sparse_vector is None:
            return None
        indices = []
        values = []
        for k, v in sparse_vector.items():
            indices.append(k)
            values.append(v)
        return models.SparseVector(indices=indices, values=values)

    def convert_expr(self, expr: Expression) -> Any:
        if expr is None or isinstance(
            expr, (str, int, float, bool, dict, list)
        ):
            return expr
        if isinstance(expr, Field):
            field = self.processor.resolve_metadata_field(expr.path)
            return field
        if isinstance(expr, Function):
            return self.convert_func(expr)
        if isinstance(expr, Comparison):
            op = expr.op
            if isinstance(expr.lexpr, Field):
                field = self.convert_expr(expr.lexpr)
                value = self.convert_expr(expr.rexpr)
            elif isinstance(expr.rexpr, Field):
                field = self.convert_expr(expr.rexpr)
                value = self.convert_expr(expr.lexpr)
                op = Comparison.reverse_op(op)
            if op == ComparisonOp.EQ:
                return models.FieldCondition(
                    key=field, match=models.MatchValue(value=value)
                )
            elif op == ComparisonOp.NEQ:
                return models.Filter(
                    must_not=[
                        models.FieldCondition(
                            key=field, match=models.MatchValue(value=value)
                        )
                    ]
                )
            elif op == ComparisonOp.GT:
                return models.FieldCondition(
                    key=field, range=models.Range(gt=value)
                )
            elif op == ComparisonOp.GTE:
                return models.FieldCondition(
                    key=field, range=models.Range(gte=value)
                )
            elif op == ComparisonOp.LT:
                return models.FieldCondition(
                    key=field, range=models.Range(lt=value)
                )
            elif op == ComparisonOp.LTE:
                return models.FieldCondition(
                    key=field, range=models.Range(lte=value)
                )
            elif op == ComparisonOp.BETWEEN:
                return models.Filter(
                    must=[
                        models.FieldCondition(
                            key=field, range=models.Range(gte=value[0])
                        ),
                        models.FieldCondition(
                            key=field, range=models.Range(lte=value[1])
                        ),
                    ]
                )
            elif op == ComparisonOp.IN:
                return models.FieldCondition(
                    key=field, match=models.MatchAny(any=value)
                )
            elif op == ComparisonOp.NIN:
                return models.FieldCondition(
                    key=field, match=models.MatchExcept(**{"except": value})
                )
            else:
                raise BadRequestError(f"Comparison op {expr.op} not supported")
        if isinstance(expr, And):
            return models.Filter(
                must=[
                    self.convert_expr(expr.lexpr),
                    self.convert_expr(expr.rexpr),
                ]
            )
        if isinstance(expr, Or):
            return models.Filter(
                should=[
                    self.convert_expr(expr.lexpr),
                    self.convert_expr(expr.rexpr),
                ]
            )
        if isinstance(expr, Not):
            return models.Filter(must_not=[self.convert_expr(expr.expr)])
        return str(expr)

    def convert_func(self, expr: Function) -> Any:
        namespace = expr.namespace
        name = expr.name
        args = expr.args
        if namespace == FunctionNamespace.BUILTIN:
            if name == StoreFunctionName.IS_DEFINED:
                field = self.convert_expr(args[0])
                return models.Filter(
                    must_not=[
                        models.IsEmptyCondition(
                            is_empty=models.PayloadField(key=field)
                        )
                    ]
                )
            if name == StoreFunctionName.IS_NOT_DEFINED:
                field = self.convert_expr(args[0])
                return models.IsEmptyCondition(
                    is_empty=models.PayloadField(key=field)
                )
            if name == StoreFunctionName.ARRAY_CONTAINS:
                field = self.convert_expr(args[0])
                return models.FieldCondition(
                    key=field, match=models.MatchValue(value=args[1])
                )
            if name == StoreFunctionName.ARRAY_CONTAINS_ANY:
                field = self.convert_expr(args[0])
                return models.FieldCondition(
                    key=field, match=models.MatchAny(any=args[1])
                )
        raise BadRequestError(f"Function {name} not supported")

    def convert_count(self, where: Expression) -> dict | None:
        args: dict = {"collection_name": self.collection}
        if where is not None:
            args["count_filter"] = models.Filter(
                must=[self.convert_expr(where)]
            )
        return args


class QdrantCollection:
    client: Any
    converter: OperationConverter
    processor: ItemProcessor
    vector_field: str
    sparse_vector_field: str | None

    def __init__(
        self,
        client: Any,
        collection: str | None,
        dimension: int,
        vector_field: str,
        sparse_vector_field: str | None,
        content_payload_field: str | None,
    ):
        self.client = client
        self.processor = ItemProcessor()
        self.converter = OperationConverter(
            processor=self.processor,
            collection=collection,
            dimension=dimension,
            vector_field=vector_field,
            sparse_vector_field=sparse_vector_field,
            content_payload_field=content_payload_field,
        )



from __future__ import annotations

from typing import Any

from verse.core import Operation
from verse.ql import Collection, Expression, OrderBy, Select, Update, Value


class StoreOperation:
    EXECUTE = "execute"
    GET = "get"
    PUT = "put"
    UPDATE = "update"
    DELETE = "delete"
    QUERY = "query"
    COUNT = "count"
    BATCH = "batch"
    TRANSACT = "transact"
    COPY = "copy"
    GENERATE = "generate"
    WATCH = "watch"
    CLOSE = "close"

    CREATE_COLLECTION = "create_collection"
    DROP_COLLECTION = "drop_collection"
    LIST_COLLECTIONS = "list_collections"

    CREATE_INDEX = "create_index"
    DROP_INDEX = "drop_index"
    LIST_INDEXES = "list_indexes"

    @staticmethod
    def execute(
        statement: str,
        params: dict[str, Any] | None = None,
        **kwargs,
    ):
        return Operation.normalize(StoreOperation.EXECUTE, locals())

    @staticmethod
    def create_collection(
        collection: str | Collection | None = None,
        **kwargs,
    ):
        return Operation.normalize(StoreOperation.CREATE_COLLECTION, locals())

    @staticmethod
    def drop_collection(
        collection: str | Collection | None = None,
        **kwargs,
    ):
        return Operation.normalize(StoreOperation.DROP_COLLECTION, locals())

    @staticmethod
    def list_collections(
        **kwargs,
    ):
        return Operation.normalize(StoreOperation.LIST_COLLECTIONS, locals())

    @staticmethod
    def get(
        key: Any = None,
        where: str | Expression | None = None,
        collection: str | Collection | None = None,
        **kwargs,
    ):
        return Operation.normalize(StoreOperation.GET, locals())

    @staticmethod
    def put(
        key: Any = None,
        value: Value | None = None,
        where: str | Expression | None = None,
        collection: str | Collection | None = None,
        **kwargs,
    ):
        return Operation.normalize(StoreOperation.PUT, locals())

    @staticmethod
    def update(
        key: Any = None,
        set: str | Update | None = None,
        where: str | Expression | None = None,
        returning: bool | None = None,
        collection: str | Collection | None = None,
        **kwargs,
    ):
        return Operation.normalize(StoreOperation.UPDATE, locals())

    @staticmethod
    def delete(
        key: Any = None,
        where: str | Expression | None = None,
        collection: str | Collection | None = None,
        **kwargs,
    ):
        return Operation.normalize(StoreOperation.DELETE, locals())

    @staticmethod
    def query(
        select: str | Select | None = None,
        search: str | Expression | None = None,
        where: str | Expression | None = None,
        order_by: str | OrderBy | None = None,
        limit: int | None = None,
        offset: int | None = None,
        collection: str | Collection | None = None,
        **kwargs,
    ):
        return Operation.normalize(StoreOperation.QUERY, locals())

    @staticmethod
    def count(
        search: str | Expression | None = None,
        where: str | Expression | None = None,
        collection: str | Collection | None = None,
        **kwargs,
    ):
        return Operation.normalize(StoreOperation.COUNT, locals())

    @staticmethod
    def batch(
        operations: list[Operation],
        **kwargs,
    ):
        return Operation.normalize(StoreOperation.BATCH, locals())

    @staticmethod
    def transact(
        operations: list[Operation],
        **kwargs,
    ):
        return Operation.normalize(StoreOperation.TRANSACT, locals())

    @staticmethod
    def copy(
        key: Any = None,
        source: Value = None,
        where: str | Expression | None = None,
        collection: str | Collection | None = None,
        **kwargs,
    ):
        return Operation.normalize(StoreOperation.COPY, locals())

    @staticmethod
    def generate(
        key: Any = None,
        collection: str | Collection | None = None,
        **kwargs,
    ):
        return Operation.normalize(StoreOperation.GENERATE, locals())

    @staticmethod
    def close(**kwargs):
        return Operation.normalize(StoreOperation.CLOSE, locals())
